package acceptance;

import java.io.PrintStream;
import java.util.BitSet;

import acceptance.AcceptanceRabin.RabinPair;
import jdd.JDDVars;
import prism.PrismException;
import prism.PrismNotSupportedException;

public class AcceptanceCoBuchi implements AcceptanceOmega {
	/** The set of goal states */
	private BitSet acceptingStates = new BitSet();

	/** Constructor (no accepting states) */
	public AcceptanceCoBuchi()
	{
	}

	/** Constructor (set accepting states) */
	public AcceptanceCoBuchi(BitSet acceptingStates)
	{
		this.acceptingStates = acceptingStates;
	}

	/** Get the accepting state set */
	public BitSet getAcceptingStates()
	{
		return acceptingStates;
	}

	/** Set the accepting state set */
	public void setAcceptingStates(BitSet acceptingStates)
	{
		this.acceptingStates = acceptingStates;
	}

	/** Make a copy of the acceptance condition. */
	public AcceptanceCoBuchi clone()
	{
		return new AcceptanceCoBuchi((BitSet)acceptingStates.clone());
	}
	
	@Override
	public boolean isBSCCAccepting(BitSet bscc_states)
	{
		return !bscc_states.intersects(acceptingStates);
	}

	/**
	 * Get the Rabin acceptance condition that is the equivalent of this Buchi condition.
	 */
	public AcceptanceRabin toRabin(int numStates)
	{
		AcceptanceRabin rabin = new AcceptanceRabin();
		RabinPair pair = new AcceptanceRabin.RabinPair((BitSet) acceptingStates.clone(), null);
		pair.setFinUsed(true);
		pair.setInfUsed(false);
		rabin.add(pair);
		return rabin;
	}

	/**
	 * Get a Rabin acceptance condition that is the complement of this condition, i.e.,
	 * any word that is accepted by this condition is rejected by the returned Rabin condition.
	 *
	 * @param numStates the number of states in the underlying model / automaton (needed for complementing BitSets)
	 * @return the complement Rabin acceptance condition
	 */
	public AcceptanceRabin complementToRabin(int numStates)
	{
		AcceptanceRabin rabin = new AcceptanceRabin();
		BitSet allStates = new BitSet();
		allStates.set(0, numStates);
		RabinPair pair = new AcceptanceRabin.RabinPair((BitSet) acceptingStates.clone(), null);
		pair.setFinUsed(true);
		pair.setInfUsed(false);
		rabin.add(pair);
		return rabin;
	}

	/**
	 * Get a Streett acceptance condition that is the complement of this condition, i.e.,
	 * any word that is accepted by this condition is rejected by the returned Streett condition.
	 * <br>
	 * Relies on the fact that once the goal states have been reached, all subsequent states
	 * are goal states.
	 *
	 * @param numStates the number of states in the underlying model / automaton (needed for complementing BitSets)
	 * @return the complement Streett acceptance condition
	 */
	public AcceptanceStreett complementToStreett(int numStates)
	{
		AcceptanceStreett streett = new AcceptanceStreett();
		streett.add(new AcceptanceStreett.StreettPair((BitSet) acceptingStates.clone(), new BitSet()));
		return streett;
	}

	/** Complement this acceptance condition, return as AcceptanceGeneric. */
	public AcceptanceGeneric complementToGeneric()
	{
		return toAcceptanceGeneric().complementToGeneric();
	}

	@Override
	public AcceptanceOmega complement(int numStates, AcceptanceType... allowedAcceptance) throws PrismException
	{
		if (AcceptanceType.contains(allowedAcceptance, AcceptanceType.RABIN)) {
			return complementToRabin(numStates);
		} else if (AcceptanceType.contains(allowedAcceptance, AcceptanceType.STREETT)) {
			return complementToStreett(numStates);
		} else if (AcceptanceType.contains(allowedAcceptance, AcceptanceType.GENERIC)) {
			return complementToGeneric();
		}
		throw new PrismNotSupportedException("Can not complement " + getType() + " acceptance to a supported acceptance type");
	}

	@Override
	public void lift(LiftBitSet lifter)
	{
		acceptingStates=lifter.lift(acceptingStates);
	}

	@Override
	public AcceptanceOmegaDD toAcceptanceDD(JDDVars ddRowVars)
	{
		// TODO change to AcceptanceCoBuchiDD
		return toAcceptanceGeneric().toAcceptanceDD(ddRowVars);
	}

	@Override
	public AcceptanceGeneric toAcceptanceGeneric()
	{
		return new AcceptanceGeneric(AcceptanceGeneric.ElementType.INF, (BitSet) acceptingStates.clone());
	}

	@Override
	public String getSignatureForState(int i)
	{
		return acceptingStates.get(i) ? "!" : " ";
	}
	
	@Override
	public String getSignatureForStateHOA(int stateIndex)
	{
		if (acceptingStates.get(stateIndex)) {
			return "{0}";
		} else {
			return "";
		}
	}

	/** Returns a textual representation of this acceptance condition. */
	@Override
	public String toString()
	{
		return acceptingStates.toString();
	}

	@Override
	public String getSizeStatistics()
	{
		return acceptingStates.cardinality()+" accepting states";
	}

	@Override
	public AcceptanceType getType()
	{
		return AcceptanceType.BUCHI;
	}

	@Override
	@Deprecated
	public String getTypeAbbreviated()
	{
		return getType().getNameAbbreviated();
	}

	@Override
	@Deprecated
	public String getTypeName()
	{
		return getType().getName();
	}

	@Override
	public void outputHOAHeader(PrintStream out)
	{
		out.println("acc-name: co-Buchi");
		out.println("Acceptance: 1 Fin(0)");
	}

}
