package acceptance;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;
import java.util.stream.Collectors;

import acceptance.AcceptanceGeneric.ElementType;
import jdd.JDDVars;
import prism.PrismNotSupportedException;

public class AcceptanceGenBuchi implements AcceptanceOmega {
	/** The set of goal states */
	private List<BitSet> acceptingStates = new ArrayList<BitSet>();

	/** Constructor (no accepting states) */
	public AcceptanceGenBuchi()
	{
	}

	/** Constructor (set accepting states) */
	public AcceptanceGenBuchi(List<BitSet> acceptingStates)
	{
		this.acceptingStates = acceptingStates;
	}

	/** Get the accepting state set */
	public List<BitSet> getAcceptingStates()
	{
		return acceptingStates;
	}

	/** Set the accepting state set */
	public void setAcceptingStates(List<BitSet> acceptingStates)
	{
		this.acceptingStates = acceptingStates;
	}

	/** Make a copy of the acceptance condition. */
	public AcceptanceGenBuchi clone()
	{
		return new AcceptanceGenBuchi(new ArrayList<BitSet>(acceptingStates));
	}
	
	@Override
	public boolean isBSCCAccepting(BitSet bscc_states)
	{
		for(BitSet accSet : acceptingStates) {
			if(!(bscc_states.intersects(accSet))) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Get the Rabin acceptance condition that is the equivalent of this Buchi condition.
	 */
	public AcceptanceGenRabin toRabin()
	{
		AcceptanceGenRabin rabin = new AcceptanceGenRabin();
		AcceptanceGenRabin.GenRabinPair pair = 
				new AcceptanceGenRabin.GenRabinPair(null, 
						new ArrayList<BitSet>(acceptingStates));
		pair.setFinUsed(false);
		pair.setInfUsed(true);
		rabin.add(pair);
		return rabin;
	}


	@Override
	public void lift(LiftBitSet lifter)
	{
		acceptingStates = acceptingStates.stream().
				map(stateSet -> lifter.lift(stateSet)).
				collect(Collectors.toList());
	}

	@Override
	public AcceptanceGenBuchiDD toAcceptanceDD(JDDVars ddRowVars)
	{
		return new AcceptanceGenBuchiDD(this, ddRowVars);
	}

	@Override
	public AcceptanceGeneric toAcceptanceGeneric()
	{
		if(acceptingStates.isEmpty()) {
			return new AcceptanceGeneric(true);
		}
		
		boolean first = true;
		AcceptanceGeneric result = null;
		for(BitSet stateSet : acceptingStates) {
			AcceptanceGeneric buchi = new AcceptanceGeneric(ElementType.INF, stateSet);
			if(first) {
				result = buchi;
				first = false;
			} else {
				result = new AcceptanceGeneric(ElementType.AND,
						result,
						buchi);
			}
		}
		
		return result;
	}

	@Override
	public String getSignatureForState(int i)
	{
		String result = "";
		for(int k = 0; k < acceptingStates.size(); k++) {
			String comma = k==0 ? "" : ",";
			result += acceptingStates.get(k).get(i) ? (comma + "!" + i) : ""; 
		}
		return result;
	}
	
	@Override
	public String getSignatureForStateHOA(int stateIndex)
	{
		if(acceptingStates.isEmpty()) {
			return "";
		}
		String result = "{";
		for(int i = 0; i < acceptingStates.size(); i++) {
			if(acceptingStates.get(i).get(stateIndex)) {
				result += 
						(result.equals("{") ? i : "," + i);
			}
		}
		return result + "}";
	}

	/** Returns a textual representation of this acceptance condition. */
	@Override
	public String toString()
	{
		return acceptingStates.toString();
	}

	@Override
	public String getSizeStatistics()
	{
		return acceptingStates.size()+" accepting sets";
	}

	@Override
	public AcceptanceType getType()
	{
		return AcceptanceType.BUCHI;
	}

	@Override
	@Deprecated
	public String getTypeAbbreviated()
	{
		return getType().getNameAbbreviated();
	}

	@Override
	@Deprecated
	public String getTypeName()
	{
		return getType().getName();
	}

	@Override
	public void outputHOAHeader(PrintStream out)
	{
		out.println("acc-name: Buchi");
		out.println("Acceptance: 1 Inf(0)");
	}

	public AcceptanceOmega complement(int numStates, AcceptanceType... allowedAcceptance) throws PrismNotSupportedException
	{
		if (AcceptanceType.contains(allowedAcceptance, AcceptanceType.RABIN)) {
			return complementToRabin();
		} else if (AcceptanceType.contains(allowedAcceptance, AcceptanceType.GENERIC)) {
			return complementToGeneric();
		}
		throw new PrismNotSupportedException("Can not complement " + getType() + " acceptance to a supported acceptance type");
	}

	private AcceptanceGeneric complementToGeneric()
	{
		if(acceptingStates.isEmpty()) {
			return new AcceptanceGeneric(false);
		}
		
		boolean first = true;
		AcceptanceGeneric result = null;
		for(BitSet accSet : acceptingStates) {
			if(first) {
				result = new AcceptanceGeneric(ElementType.FIN, accSet);
				first = false;
			} else {
				result = result.or(new AcceptanceGeneric(ElementType.FIN, accSet));
			}
		}
		return result;
	}

	/**
	 * Get a Rabin acceptance condition that is the complement of this condition, i.e.,
	 * any word that is accepted by this condition is rejected by the returned Rabin condition.
	 *
	 * @return the complement Rabin acceptance condition
	 */
	public AcceptanceRabin complementToRabin()
	{
		AcceptanceRabin rabin = new AcceptanceRabin();
		for(BitSet stateSet : acceptingStates) {
			AcceptanceRabin.RabinPair pair =
					new AcceptanceRabin.RabinPair((BitSet)(stateSet.clone()), null);
			pair.setFinUsed(true);
			pair.setInfUsed(false);
			rabin.add(pair);
		}
		return rabin;
	}
}
