package acceptance;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;
import java.util.stream.Collectors;

import acceptance.AcceptanceGeneric.ElementType;
import common.IterableBitSet;
import jdd.JDD;
import jdd.JDDNode;
import jdd.JDDVars;
import prism.PrismNotSupportedException;

/**
 * A Büchi acceptance condition (based on JDD state sets).
 * The acceptance is defined via a set of "accepting" states
 * (sometimes also called final states) and is accepting if
 *  "infinitely often an accepting state is visited"
 */
public class AcceptanceGenBuchiDD implements AcceptanceOmegaDD
{
	/** The accepting states */
	private List<JDDNode> acceptingStates;

	/**
	 * Constructor, set acceptingStates.
	 * Becomes owner of the references of acceptingStates.
	 */
	public AcceptanceGenBuchiDD(List<JDDNode> acceptingStates)
	{
		this.acceptingStates = acceptingStates;
	}
	
	/**
	 * Constructor, from a BitSet-based AcceptanceBuchi.
	 *
	 * @param acceptance the BitSet-based acceptance condition
	 * @param ddRowVars JDDVars of the row variables corresponding to the bits in the bit set
	 */
	public AcceptanceGenBuchiDD(AcceptanceGenBuchi acceptance, JDDVars ddRowVars)
	{
		acceptingStates = new ArrayList<>(acceptance.getAcceptingStates().size());
		for(BitSet accSet : acceptance.getAcceptingStates()) {
			JDDNode buchi = JDD.Constant(0);
			// get BDD based on the acceptingState bit set
			for (int i : IterableBitSet.getSetBits(accSet)) {
				buchi = JDD.SetVectorElement(buchi, ddRowVars, i, 1.0);
			}
			acceptingStates.add(buchi);
		}
	}

	/** Get a referenced copy of the state set of the accepting states.
	 * <br>[ REFS: <i>result</i>, DEREFS: <i>none</i> ]
	 */
	public List<JDDNode> getAcceptingStates()
	{
		for(JDDNode stateSet : acceptingStates) {
			JDD.Ref(stateSet);
		}
		return acceptingStates;
	}

	/**
	 * Set the accepting states.
	 * Becomes owner of the reference to acceptingStates.
	 */
	public void setAcceptingStates(List<JDDNode> acceptingStates)
	{
		clear();
		this.acceptingStates = acceptingStates;
	}

	@Override
	public boolean isBSCCAccepting(JDDNode bscc_states)
	{
		for(JDDNode stateSet : acceptingStates) {
			if(!JDD.AreIntersecting(stateSet, bscc_states)) {
				return false;
			}
		}
		return true;
	}

	@Override
	public AcceptanceGenBuchiDD clone()
	{
		return new AcceptanceGenBuchiDD(new ArrayList<JDDNode>(acceptingStates));
	}

	@Override
	public void intersect(JDDNode restrict)
	{
		acceptingStates = 
				acceptingStates.stream().
					map(stateSet -> JDD.And(stateSet, restrict.copy())).
					collect(Collectors.toList());
	}

	@Override
	public String getSizeStatistics()
	{
		return "one set of accepting states";
	}

	@Override
	public AcceptanceType getType()
	{
		return AcceptanceType.BUCHI;
	}


	@Override
	public AcceptanceOmegaDD complement(AcceptanceType... allowedAcceptance) throws PrismNotSupportedException
	{
		if (AcceptanceType.contains(allowedAcceptance, AcceptanceType.RABIN)) {
			return complementToRabin();
		} else if (AcceptanceType.contains(allowedAcceptance, AcceptanceType.GENERIC)) {
			return complementToGeneric();
		}
		throw new PrismNotSupportedException("Can not complement " + getType() + " acceptance to a supported acceptance type");
	}

	/**
	 * Get a Rabin acceptance condition that is the complement of this condition, i.e.,
	 * any word that is accepted by this condition is rejected by the returned Rabin condition.
	 *
	 * @return the complement Rabin acceptance condition
	 */
	public AcceptanceRabinDD complementToRabin()
	{
		AcceptanceRabinDD rabin = new AcceptanceRabinDD();
		for(JDDNode stateSet : acceptingStates) {
			rabin.add(new AcceptanceRabinDD.RabinPairDD(stateSet.copy(), JDD.Constant(1)));
		}
		return rabin;
	}

	@Override
	public AcceptanceGenericDD toAcceptanceGeneric()
	{	
		if(acceptingStates.isEmpty()) {
			return new AcceptanceGenericDD(true);
		}
		
		boolean first = true;
		AcceptanceGenericDD result = null;
		for(JDDNode stateSet : acceptingStates) {
			AcceptanceGenericDD buchi = new AcceptanceGenericDD(ElementType.INF, stateSet);
			if(first) {
				result = buchi;
				first = false;
			} else {
				result = new AcceptanceGenericDD(ElementType.AND,
						result,
						buchi);
			}
		}
		
		return result;
	}

	@Override
	@Deprecated
	public String getTypeAbbreviated()
	{
		return getType().getNameAbbreviated();
	}

	@Override
	@Deprecated
	public String getTypeName()
	{
		return getType().getName();
	}

	@Override
	public void clear()
	{
		for(JDDNode stateSet : acceptingStates) {
			JDD.Deref(stateSet);
		}
	}

}
