package explicit;

import java.util.BitSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import acceptance.AcceptanceGeneric;
import prism.PrismException;

public class GenericEmptiness {
	
	private static	Map<AcceptanceGeneric,Integer> acc2SlackVar = new LinkedHashMap<AcceptanceGeneric,Integer>();
	private MDPModelChecker mc;
	private MDP mdp;
	
	protected static Map<AcceptanceGeneric,Integer> getAcc2SlackVar()
	{
		return acc2SlackVar;
	}
	
	protected static int findOrAdd(AcceptanceGeneric acc) throws PrismException
	{
		Map<AcceptanceGeneric,Integer> acc2SlackVar = getAcc2SlackVar();
		if(!acc2SlackVar.containsKey(acc)) {
			int max = acc2SlackVar.size();
			
			//Sanity check: every variable should be inserted only once
			if(acc2SlackVar.containsValue(max)) {
				throw new PrismException("Integer " + max + " is contained two times in hash table for the Tseytin transformation of " + acc);
			}
			
			acc2SlackVar.put(acc, max);
		}
		return acc2SlackVar.get(acc);
	}

	public enum APKind {
		INF,
		FIN,
		SLACKVAR,
		NEGSLACKVAR
	}
	
	class AP 
	{
		APKind kind;
		BitSet states;
		int varIndex;
		
		public AP(int varIndex, boolean negated) 
		{
			kind = (negated==true) ? APKind.NEGSLACKVAR:APKind.SLACKVAR;
			this.varIndex = varIndex;
		}
		
		public AP(AcceptanceGeneric acc) throws PrismException
		{
			switch(acc.getKind()) {
			case FIN:
				kind = APKind.FIN;
				states = (BitSet)acc.getStates().clone();
				break;
			case INF:
				kind = APKind.INF;
				states = (BitSet)acc.getStates().clone();
				break;
			default:
				throw new PrismException("Not implemented");
			}
		}
		
		public AP(APKind kind, BitSet states)
		{
			this.kind = kind;
			this.states = states;
		}

		public APKind getKind() {
			return kind;
		}

		public BitSet getStates() {
			return states;
		}

		public int getVarIndex() {
			return varIndex;
		}
		
		public int hashCode() {
			switch(kind) {
			case INF:
				return states.hashCode();
			case FIN:
				return ~states.hashCode();
			case SLACKVAR:
				return varIndex;
			case NEGSLACKVAR:
				return ~varIndex;
			}
			return 0;
		}
			
		public boolean equals(Object o) {
			if(o instanceof AP){
				AP otherAP = (AP)o;
				
				if(otherAP.getKind() != getKind()) {
					return false;
				}
				
				if(getKind() == APKind.INF || getKind() == APKind.FIN) {
					return otherAP.getStates().equals(getStates());
				}
				
				return otherAP.getVarIndex() == getVarIndex();
			}
			
			return false;
		}
		
		public String toString() 
		{
			switch(getKind()) {
			case SLACKVAR:
				return "S" + getVarIndex();
			case NEGSLACKVAR:
				return "!S" + getVarIndex();
			case INF:
				return "Inf(" + getStates() + ")";
			case FIN:
				return "Fin(" + getStates() + ")";
			default:
				return "Java doesn't recognizes a complete enumeration.";
			}
		}
	}
	
	class Clause extends LinkedHashSet<AP>
	{
		
		/**
		 * 
		 */
		private static final long serialVersionUID = -3981555988284552486L;
		boolean isValid = false;
		
		public boolean isTerminal() 
		{
			if(isEmpty() || isValid) {
				return true;
			}
			
			if(size() != 2) {
				return false;
			}
			
			//is true iff NEGSLACKVAR or FIN has been encountered
			boolean foundNeg = false;
			for(AP ap: this) {
				if(ap.getKind() == APKind.FIN || ap.getKind() == APKind.NEGSLACKVAR) {
					if(foundNeg) {
						return false;
					}
					foundNeg = true;
				}
			}
			
			return true;
		}
		
		public boolean isTrue() {
			return isValid;
		}
		
		public void setTrue() {
			isValid = true;
		}
		
		
		public void cutWithMEC(BitSet mec)
		{
			Iterator<AP> it = iterator();
			while(it.hasNext()) {
				AP ap = it.next();
				if(ap.getKind() == APKind.INF || ap.getKind() == APKind.FIN) {
					ap.getStates().and(mec);
				}
			}
			simplify();
		}
		
		public void simplify()
		{
			if(isValid) {
				clear();
				return;
			}
			
			//Search for empty Fin
			Iterator<AP> it = iterator();
			while(it.hasNext()) {
				AP ap = it.next();
				if(ap.getKind() == APKind.FIN && ap.getStates().isEmpty()) {
					isValid = true;
					return;
				}
			}
			
			//TODO: Set clause to true, if the are aps of the form p and not p (and analogously Fin and Inf)
			
			//TODO: Remove Fin(S'), if there is a Fin(S) with S \subseteq S'
			
			//Unite all Infs
			AP inf = new AP(APKind.INF, new BitSet());
			it = iterator();
			while(it.hasNext()) {
				AP ap = it.next();
				if(ap.getKind() == APKind.INF) {
					inf.getStates().or(ap.getStates());
					it.remove();
				}
			}
			if(!inf.getStates().isEmpty()) {
				add(inf);
			}
		}
		
		public String toString() 
		{
			if(isTrue()) {
				return "true";
			}
			
			if(isFalse()) {
				return "false";
			}
			
			String result;
			Iterator<AP> it = iterator();
			result = it.next().toString();
			while(it.hasNext()) {
				result += " | " + it.next().toString();
			}
			return result;
		}

		private boolean isFalse()
		{
			return size() == 0;
		}
	}
	
	class CNF implements Iterable<Clause> 
	{

		Set<Clause> clauses;
		
		public void cutWithMEC(BitSet mec)
		{
			for(Clause clause : clauses) {
				clause.cutWithMEC(mec);
			}
			simplify();
		}
		
		public boolean isTerminal()
		{
			if(clauses.isEmpty()) {
				return true;
			}
			
			for(Clause clause: clauses) {
				if(clause.isEmpty()) {
					return true;
				}
			}
			
			return false;
		}
		
		public CNF(Set<Clause> clauses)
		{
			this.clauses = clauses;
		}
	
		public void simplify()
		{
			if(isFalse()) {
				clauses.clear();
				clauses.add(new Clause());
				return;
			}
			if(isTrue()) {
				return;
			}
			
			Iterator<Clause> it = clauses.iterator();
			while(it.hasNext()) {
				Clause clause = it.next();
				clause.simplify();
				if(clause.isTrue()) {
					it.remove();
				}
			}
		}
		
		public Iterator<Clause> iterator() {
			return new CNFIterator(clauses);
		}
		
		private Set<Clause> createClauseSet(AcceptanceGeneric acc) throws PrismException
		{
			Set<Clause> clauses = new LinkedHashSet<Clause>();
			
			Clause clause = new Clause();
			Clause clauseLeft = new Clause();
			Clause clauseRight = new Clause();
			
			switch(acc.getKind()) {
			case FIN:
				clause.add(new AP(GenericEmptiness.APKind.FIN, acc.getStates()));
				clauses.add(clause);
				break;
			case INF:
				clause.add(new AP(GenericEmptiness.APKind.INF, acc.getStates()));
				clauses.add(clause);
				break;
			case AND:
				//Rule for and:
				//s_acc -> s_left & s_acc -> s_right
				AP negAP = new AP(findOrAdd(acc),true);
				clauseLeft.add(negAP);
				if(acc.getLeft().getKind() == AcceptanceGeneric.ElementType.FIN ||
						acc.getLeft().getKind() == AcceptanceGeneric.ElementType.INF) {
					//Add a FIN or INF directly
					clauseLeft.add(new AP(acc.getLeft()));
				} else {
					//Add a boolean combination via findOrAdd
					clauseLeft.add(new AP(findOrAdd(acc.getLeft()), false));
					clauses.addAll(createClauseSet(acc.getLeft()));
				}
				clauses.add(clauseLeft);
				
				clauseRight.add(negAP);
				if(acc.getRight().getKind() == AcceptanceGeneric.ElementType.FIN ||
						acc.getRight().getKind() == AcceptanceGeneric.ElementType.INF) {
					//Add a FIN or INF directly
					clauseRight.add(new AP(acc.getRight()));
				} else {
					//Add a boolean combination via findOrAdd
					clauseRight.add(new AP(findOrAdd(acc.getRight()), false));
					clauses.addAll(createClauseSet(acc.getRight()));
				}
				clauses.add(clauseRight);
				
				break;
			case OR:
				//Rule for or:
				//s_acc -> (s_left | s_right)
				negAP = new AP(findOrAdd(acc),true);

				clause.add(negAP);
				if(acc.getLeft().getKind() == AcceptanceGeneric.ElementType.FIN ||
						acc.getLeft().getKind() == AcceptanceGeneric.ElementType.INF) {
					//Add a FIN or INF directly
					clause.add(new AP(acc.getLeft()));
				} else {
					//Add a boolean combination via findOrAdd
					clause.add(new AP(findOrAdd(acc.getLeft()), false));
					clauses.addAll(createClauseSet(acc.getLeft()));
				}
				if(acc.getRight().getKind() == AcceptanceGeneric.ElementType.FIN ||
						acc.getRight().getKind() == AcceptanceGeneric.ElementType.INF) {
					//Add a FIN or INF directly
					clause.add(new AP(acc.getRight()));
				} else {
					//Add a boolean combination via findOrAdd
					clause.add(new AP(findOrAdd(acc.getRight()), false));
					clauses.addAll(createClauseSet(acc.getRight()));
				}
				
				clauses.add(clause);
				
				break;
			case FALSE:
			case FIN_NOT:
			case INF_NOT:
			case TRUE:
			default:
				throw new PrismException("Tseytin transformation is not yet implemented for FALSE, FIN_NOT, INF_NOT and TRUE");
			}
			
			return clauses;
		}
		
		public Set<Clause> getClauses()
		{
			return clauses;
		}
		
		public CNF(AcceptanceGeneric acc) throws PrismException {
			clauses = createClauseSet(acc);

			Clause top = new Clause();
			top.add(new AP(findOrAdd(acc),false));
			
			clauses.add(top);
		}

		public boolean isTrue()
		{
			for(Clause clause : clauses) {
				clause.simplify();
				if(clause.isFalse()) {
					return false;
				}
				if(!clause.isTrue()) {
					//CNF, that contains only INF as atomic propositions are also true 
					for(AP ap : clause) {
						if(ap.getKind() != APKind.INF) {
							return false;
						}
					}
				}
			}
			return true;
		}
		
		public boolean isFalse()
		{
			//A cnf formula is false, if it contains an empty clause
			return clauses.contains(new Clause());
		}
		
		public String toString() 
		{
			if(isTrue()) {
				return "true";
			}
			if(isFalse()) {
				return "false";
			}
			String result = "";
			Iterator<Clause> it = clauses.iterator();
			result = "(" + it.next().toString() + ")";
			while(it.hasNext()) {
				result += " & (" + it.next().toString() + ")";
			}
			return result;
		}
		
		public AP chooseAP() 
		{
			//Until now: Choose first variable
			AP chosenAP = null;
			for(Clause clause: this) {
				//TODO: Define own iterator for CNF
				if(!clause.isFalse()) {
					for(AP ap: clause) {
						if(ap.getKind() == APKind.SLACKVAR ||
								ap.getKind() == APKind.NEGSLACKVAR ||
								ap.getKind() == APKind.FIN) {
							chosenAP = ap;
							break;
						}
					}
				}
				if(chosenAP != null) {
					break;
				}
			}
			return chosenAP;
		}
		
		
	}
	
	class CNFIterator implements Iterator<Clause> 
	{

		Iterator<Clause> it;
		Clause next = null;
		
		public CNFIterator(Set<Clause> clauses)
		{
			it = clauses.iterator();
		}
		
		@Override
		public boolean hasNext() {
			if(next != null) {
				return true;
			}

			while(it.hasNext()) {
				next = it.next();
				if(!next.isTrue()) {
					return true;
				}
			}
			return false;
		}

		@Override
		public Clause next() {
			hasNext();
			Clause nextCp = next;
			next = null;
			return nextCp;
		}
		
	}
	
	public Clause filterClauseNeg(Clause clause, AP chosenAP)
	{
		Clause clauseCp = new Clause();

		if(clause.isTrue()) {
			clauseCp.setTrue();
			return clauseCp;
		}

		for(AP ap : clause) {
			if(chosenAP.getKind() == APKind.SLACKVAR) {
				if(ap.getKind() != APKind.SLACKVAR ||
						ap.getVarIndex() != chosenAP.getVarIndex()) {
					clauseCp.add(ap);
				}
			}
			if(chosenAP.getKind() == APKind.NEGSLACKVAR) {
				if(ap.getKind() != APKind.NEGSLACKVAR ||
						ap.getVarIndex() != chosenAP.getVarIndex()) {
					clauseCp.add(ap);
				}
			}
			if(chosenAP.getKind() == APKind.FIN) {
				if(ap.getKind() != APKind.FIN ||
						!bitSetSubsetEq(chosenAP.getStates(), ap.getStates())) {
					clauseCp.add(ap);
				}
			}
		}
		
		return  clauseCp;
		
	}

	public Clause filterClausePos(Clause clause, AP chosenAP)
	{
		Clause clauseCp = new Clause();
		
		if(clause.isTrue()) {
			clauseCp.setTrue();
			return clauseCp;
		}
		
		for(AP ap : clause) {
			if(chosenAP.getKind() == APKind.SLACKVAR) {
				if(ap.getKind() != APKind.NEGSLACKVAR ||
						ap.getVarIndex() != chosenAP.getVarIndex()) {
					clauseCp.add(ap);
				}
			}
			if(chosenAP.getKind() == APKind.NEGSLACKVAR) {
				if(ap.getKind() != APKind.SLACKVAR ||
						ap.getVarIndex() != chosenAP.getVarIndex()) {
					clauseCp.add(ap);
				}
			}
			if(chosenAP.getKind() == APKind.FIN) {
				if(ap.getKind() != APKind.INF ||
						!bitSetSubsetEq(ap.getStates(), chosenAP.getStates())) {
					clauseCp.add(ap);
				}
			}
		}
		
		return  clauseCp;
		
	}
	
	public boolean isTrueClauseNeg(Clause clause, AP chosenAP)
	{
		if(clause.isTrue()) {
			return true;
		}
		for(AP ap : clause) {
			if((chosenAP.getKind() == APKind.SLACKVAR && ap.getKind() == APKind.NEGSLACKVAR) ||
					(chosenAP.getKind() == APKind.NEGSLACKVAR && ap.getKind() == APKind.SLACKVAR)) {
				if(chosenAP.getVarIndex() == ap.getVarIndex()) {
					return true;
				}
			}
			
			if(chosenAP.getKind() == APKind.FIN && ap.getKind() == APKind.INF &&
					bitSetSubsetEq(chosenAP.getStates(), ap.getStates())) {
				return true;
			}
		}
		
		return false;
	}

	public boolean isTrueClausePos(Clause clause, AP chosenAP)
	{
		if(clause.isTrue()) {
			return true;
		}
		for(AP ap : clause) {
			if((chosenAP.getKind () == APKind.SLACKVAR || chosenAP.getKind() == APKind.NEGSLACKVAR) &&
					chosenAP.equals(ap)) {
				return true;
			}
			
			if(chosenAP.getKind() == APKind.FIN && ap.getKind() == APKind.FIN &&
					bitSetSubsetEq(ap.getStates(), chosenAP.getStates())) {
				return true;
			}
		}
		
		return false;
	}
	
	public CNF generateNegativeCNF(CNF cnf, AP chosenAP)
	{
		//the CNF, if the atom is set to false
		Set<Clause> falseCNF = new LinkedHashSet<Clause>();
		
		for(Clause clause: cnf.getClauses()) {
			//Take the true case
			Clause negClause = new Clause();
			negClause = filterClauseNeg(clause, chosenAP);
			if(!isTrueClauseNeg(negClause, chosenAP)) {
				falseCNF.add(negClause);
			}
		}
		
		return new CNF(falseCNF);
	}
	
	public AP getPureLiteral(CNF cnf)
	{
		LinkedHashSet<AP> pureAPs = new LinkedHashSet<AP>();
		LinkedHashSet<AP> removedAPs = new LinkedHashSet<AP>();
		for(Clause clause : cnf.getClauses()) {
			for(AP ap: clause) {
				if(!removedAPs.contains(ap)) {
					switch(ap.getKind()) {
					case SLACKVAR:
						AP negVar = new AP(ap.getVarIndex(),true);
						if(pureAPs.contains(negVar)) {
							pureAPs.remove(negVar);
							removedAPs.add(negVar);
							removedAPs.add(ap);
						} else {
							pureAPs.add(ap);
						}
						break;
					case NEGSLACKVAR:	
						negVar = new AP(ap.getVarIndex(),false);
						if(pureAPs.contains(negVar)) {
							pureAPs.remove(negVar);
							removedAPs.add(negVar);
							removedAPs.add(ap);
						} else {
							pureAPs.add(ap);
						}
						break;
					default:
						//TODO (for FINs...)
					}
				}
			}
		}
		if(!pureAPs.isEmpty()) {
			Iterator<AP> it = pureAPs.iterator();
			it.hasNext();
			return it.next();
		}
		return null;
	}
	
	public AP getUnitLiteral(CNF cnf)
	{
		for(Clause clause: cnf.getClauses()) {
			if(!clause.isTrue() && !clause.isFalse() && clause.size() == 1) {
				Iterator<AP> it = clause.iterator();
				it.hasNext();
				AP ap = it.next();
				if(ap.getKind() != APKind.INF) {
					return ap;
				}
			}
		}
		return null;
	}
	
	public CNF generatePositveCNF(CNF cnf, AP chosenAP)
	{
		//the CNF, if the atom is set to true
		Set<Clause> trueCNF = new LinkedHashSet<Clause>();
		
		for(Clause clause: cnf.getClauses()) {
			//Take the true case
			Clause posClause = new Clause();
			posClause = filterClausePos(clause, chosenAP);
			if(!isTrueClausePos(posClause, chosenAP)) {
				trueCNF.add(posClause);
			}
		}
		
		return new CNF(trueCNF);
	}

	public boolean hasAcceptingEC(CNF cnf, BitSet mec, MDP mdp) throws PrismException {
		
		if(mec.isEmpty()) {
			return false;
		}
		
		cnf.cutWithMEC(mec);
		if(cnf.isTrue()) {
			return true;
		}
		if(cnf.isFalse()) {
			return false;
		}
		
		//Unit rule
		AP unitAP = getUnitLiteral(cnf);
		if(unitAP != null) {
			System.out.println("Found unit clause " + unitAP + " in " + cnf);
			CNF trueCNF = generatePositveCNF(cnf, unitAP);
			BitSet cuttedMec = (BitSet)mec.clone();
			if(unitAP.getKind() == APKind.FIN) {
				cuttedMec.andNot(unitAP.getStates());
				cuttedMec = joinMECs(getMECs(cuttedMec, null));
			}
			return hasAcceptingEC(trueCNF,cuttedMec, mdp);
		}
		
		AP pureLiteral = getUnitLiteral(cnf);
		if(pureLiteral != null) {
			System.out.println("Found unit clause " + pureLiteral + " in " + cnf);
			CNF trueCNF = generatePositveCNF(cnf, pureLiteral);
			BitSet cuttedMec = (BitSet)mec.clone();
			if(pureLiteral.getKind() == APKind.FIN) {
				cuttedMec.andNot(pureLiteral.getStates());
				cuttedMec = joinMECs(getMECs(cuttedMec, null));
			}
			return hasAcceptingEC(trueCNF,cuttedMec, mdp);
		}

		//PureLiteralregel
		
		
		//Splitting rule
		
		AP chosenAP = cnf.chooseAP();
		
		//Do splitting
		System.out.println("Split :-(");
		CNF trueCNF = generatePositveCNF(cnf, chosenAP);
		BitSet cuttedMec = (BitSet)mec.clone();
		if(chosenAP.getKind() == APKind.FIN) {
			cuttedMec.andNot(chosenAP.getStates());
			cuttedMec = joinMECs(getMECs(cuttedMec, null));
		}
		
		CNF falseCNF = generateNegativeCNF(cnf, chosenAP);
		
		return hasAcceptingEC(trueCNF, cuttedMec, mdp) || hasAcceptingEC(falseCNF, mec, mdp);
	}
	
	public boolean hasAcceptingEC(AcceptanceGeneric acc, BitSet mec, MDP mdp) throws PrismException
	{
		
		CNF cnf = new CNF(acc);
		return hasAcceptingEC(cnf, mec, mdp);
	}
	
	public boolean bitSetSubsetEq(BitSet s1, BitSet s2) 
	{
		BitSet s1clone = (BitSet)s1.clone();
		s1clone.andNot(s2);
		return s1clone.isEmpty();
	}

	/**
	 * Compute the MECs within a given state space
	 * @param restrict the state space within the MECs are collected
	 * @param accept collect only MECs, that have states in accept. If null, all MECs are collected
	 * @return List of MECs 
	 * @throws PrismException
	 */
	private List<BitSet> getMECs(BitSet restrict, BitSet accept) throws PrismException 
	{
		ECComputer ecComputer = ECComputer.createECComputer(mc, mdp);
		ecComputer.computeMECStates(restrict, accept);
		return ecComputer.getMECStates();
		
	}
	
	
	public GenericEmptiness(MDPModelChecker mc, MDP mdp)
	{
		this.mc = mc;
		this.mdp = mdp;
	}
	
	/**
	 * Joins a list of MECs to a BitSet
	 * @param mecs list of MECs
	 * @return joined set of states in mecs
	 */
	private BitSet joinMECs(List<BitSet> mecs) 
	{
		BitSet joined = new BitSet();
		for(BitSet mec : mecs) {
			joined.or(mec);
		} 
		return joined;
	}
	
}
