//	* Dave Parker <d.a.parker@cs.bham.ac.uk> (University of Birmingham/Oxford)
//  * David Müller <david.mueller2@tu-dresden.de> (TU Dresden)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package explicit.satsolving;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import acceptance.AcceptanceGeneric;
import acceptance.AcceptanceOmega;
import acceptance.AcceptanceOmegaDD;
import acceptance.AcceptanceType;
import jdd.JDDVars;
import prism.PrismException;
import prism.PrismNotSupportedException;

/**
 * A generic acceptance condition (based on BitSet state sets).
 * This is an AST of a boolean formula (conjunction and disjunction) over
 * atoms of the form Inf(states), Inf(!states), Fin(states) and Fin(!states).
 * <br>
 * Semantics:
 *  Inf(states)  <=> G F states
 *  Inf(!states) <=> G F !states
 *  Fin(states)  <=> F G !states
 *  Fin(!states) <=> F G states
 */
public class AcceptanceELSummarized implements AcceptanceOmega {
	
	/** The types of elements in the AST */
	public enum ElementType {
		FALSE,
		TRUE, 

		BIGOR,
		BIGAND,

		INF,
		FIN,
		INF_NOT,
		FIN_NOT;
	}
	
	/** The type of this node in the AST */
	private ElementType kind;

	/** The childs (if it exist) */ 
	private List<AcceptanceELSummarized> childs = new LinkedList<>();

	/** The set of states (if this is one of INF, FIN, INF_NOT, FIN_NOT) */
	private BitSet states = null;

	/**
	 * Constructor for TRUE or FALSE
	 * @param value true or false?
	 */
	public AcceptanceELSummarized(AcceptanceGeneric acc) {
		switch(acc.getKind()) {
		case FALSE:
			kind = ElementType.FALSE;
			break;
		case TRUE:
			kind = ElementType.TRUE;
			break;
		case FIN:
			kind = ElementType.FIN;
			states = acc.getStates();
			break;
		case FIN_NOT:
			kind = ElementType.FIN_NOT;
			states = acc.getStates();
			break;
		case INF:
			kind = ElementType.INF;
			states = acc.getStates();
			break;
		case INF_NOT:
			kind = ElementType.INF_NOT;
			states = acc.getStates();
			break;
		case AND:
			kind = ElementType.BIGAND;
			AcceptanceELSummarized left = new AcceptanceELSummarized(acc.getLeft());
			AcceptanceELSummarized right = new AcceptanceELSummarized(acc.getRight());

			if(left.getKind() == ElementType.BIGAND) {
				childs.addAll(left.getChilds());
			} else {
				childs.add(left);
			}

			if(right.getKind() == ElementType.BIGAND) {
				childs.addAll(right.getChilds());
			} else {
				childs.add(right);
			}
			
			break;
		case OR:
			kind = ElementType.BIGOR;
			left = new AcceptanceELSummarized(acc.getLeft());
			right = new AcceptanceELSummarized(acc.getRight());

			if(left.getKind() == ElementType.BIGOR) {
				childs.addAll(left.getChilds());
			} else {
				childs.add(left);
			}

			if(right.getKind() == ElementType.BIGOR) {
				childs.addAll(right.getChilds());
			} else {
				childs.add(right);
			}
			
			break;
		default:
			break;
		}
	}

	/**
	 * Constructor for TRUE or FALSE
	 * @param value true or false?
	 */
	public AcceptanceELSummarized(boolean value) {
		kind = value ? ElementType.TRUE : ElementType.FALSE;
	}
	
	/** 
	 * Constructor for an INF, FIN, INF_NOT or FIN_NOT element. 
	 */
	public AcceptanceELSummarized(ElementType kind, BitSet states) {
		this.kind = kind;

		this.states = states;
	}

	/**
	 * Constructor for a binary operator (AND/OR).
	 * @param kind
	 * @param left
	 * @param right
	 */
	public AcceptanceELSummarized(ElementType kind, List<AcceptanceELSummarized> childs) {
		assert(kind == ElementType.BIGAND || kind == ElementType.BIGOR);
		this.kind = kind;
		
		for(AcceptanceELSummarized child: childs) {
			if(child.getKind() == kind) {
				this.childs.addAll(child.getChilds());
			} else {
				this.childs.add(child);
			}
		}
	}

	public List<AcceptanceELSummarized> getChilds() {
		return childs;
	}

	/** Get the ElementType of this AST element */
	public ElementType getKind() {
		return kind;
	}

	/** Get the state set of this element (if kind is one of INF, FIN, INF_NOT, FIN_NOT).
	 */
	public BitSet getStates() {
		return states;
	}

	@Override
	public boolean isBSCCAccepting(BitSet bscc) {
		switch(kind) {
			case TRUE: return true;
			case FALSE: return false;
			case BIGAND:
				return childs.stream().allMatch(child -> child.isBSCCAccepting(bscc));
			case BIGOR:
				return childs.stream().anyMatch(child -> child.isBSCCAccepting(bscc));
			case INF:
				// bscc |= G F states?
				// there exists a state in bscc and states
				return bscc.intersects(states);
			case INF_NOT: {
				// bscc_state |= G F !states?
				// the BSCC does not consist only of states
				BitSet bs = (BitSet) bscc.clone();
				bs.andNot(states);
				return !bs.isEmpty();
			}
			case FIN: {
				// bscc |= F G !states?
				// <=> there exists no states state in BSCC
				return !bscc.intersects(states);
			}
			case FIN_NOT: {
				// bscc |= F G states?
				// the BSCC consists entirely of states
				BitSet bs = (BitSet) bscc.clone();
				bs.and(states);
				return bs.equals(bscc);
			}
		}
		return false;
	}
	
	public AcceptanceELSummarized intersect(BitSet stateset, boolean complement)
	{
		switch(kind) {
		case TRUE:
			return new AcceptanceELSummarized(true);
		case FALSE:
			return new AcceptanceELSummarized(false);
		case BIGAND:
			List<AcceptanceELSummarized> intersectedChilds = new LinkedList<>();
			for(AcceptanceELSummarized child : childs) {
				AcceptanceELSummarized intersected = child.intersect(stateset, complement);
				if(intersected.getKind() == ElementType.FALSE) {
					return new AcceptanceELSummarized(false);
				}
				if(intersected.getKind() != ElementType.TRUE) {
					intersectedChilds.add(intersected);
				}
			}

			if(intersectedChilds.isEmpty()) {
				return new AcceptanceELSummarized(true);
			}

			return new AcceptanceELSummarized(ElementType.BIGAND, intersectedChilds);
		case BIGOR:
			intersectedChilds = new ArrayList<>();
			for(AcceptanceELSummarized child : childs) {
				AcceptanceELSummarized intersected = child.intersect(stateset, complement);
				if(intersected.getKind() == ElementType.TRUE) {
					return new AcceptanceELSummarized(true);
				}
				if(intersected.getKind() != ElementType.FALSE) {
					intersectedChilds.add(intersected);
				}
			}

			if(intersectedChilds.isEmpty()) {
				return new AcceptanceELSummarized(false);
			}

			return new AcceptanceELSummarized(ElementType.BIGOR, intersectedChilds);
		case FIN:
			BitSet intersected = (BitSet)states.clone();
			if(complement) {
				intersected.andNot(stateset);
			} else {
				intersected.and(stateset);
			}
			if(intersected.isEmpty()) {
				return new AcceptanceELSummarized(true);
			}
			return new AcceptanceELSummarized(ElementType.FIN, intersected);
		case FIN_NOT:
			intersected = (BitSet)stateset.clone();
			if(complement) {
				return null;
			}
			intersected.andNot(states);
			if(intersected.isEmpty()) {
				return new AcceptanceELSummarized(true);
			}
			return new AcceptanceELSummarized(ElementType.FIN, intersected);
		case INF:
			intersected = (BitSet)states.clone();
			if(complement) {
				intersected.andNot(stateset);
			} else {
				intersected.and(stateset);
			}
			if(intersected.isEmpty()) {
				return new AcceptanceELSummarized(false);
			}
			return new AcceptanceELSummarized(ElementType.INF, intersected);
		case INF_NOT:
			intersected = (BitSet)stateset.clone();
			if(complement) {
				return null;
			}
			intersected.andNot(states);
			if(intersected.isEmpty()) {
				return new AcceptanceELSummarized(false);
			}
			return new AcceptanceELSummarized(ElementType.INF, intersected);
		default:
			return null;
		}
	}
	
	/**
	 * Checks, whether this acceptance is true, if every Inf/Inf_Not is replaced by true
	 * @return true if this acceptance can be made true by replacing every Inf(•) with true
	 * @throws PrismNotSupportedException 
	 */
	public boolean isSatisfiedByInfs() throws PrismNotSupportedException
	{
		switch(kind) {
		case TRUE:
			return true;
		case FALSE:
			return false;
		case BIGAND:
			for(AcceptanceELSummarized child: childs) {
				if(!child.isSatisfiedByInfs()) {
					return false;
				}
			}
			return true;
		case BIGOR:
			for(AcceptanceELSummarized child: childs) {
				if(child.isSatisfiedByInfs()) {
					return true;
				}
			}
			return false;
		case FIN:
		case FIN_NOT:
			return false;
		case INF:
		case INF_NOT:
			return true;
		default:
			throw new PrismNotSupportedException("Unknown kind.");
		}
		
	}

	/** Get a list of all the (non-true/false) leaf nodes in this acceptance condition */
	public List<AcceptanceELSummarized> getLeafNodes()
	{
		switch (getKind()) {
		case BIGAND:
		case BIGOR: {
			List<AcceptanceELSummarized> result = new ArrayList<>();
			for(AcceptanceELSummarized child: childs) {
				result.addAll(child.getLeafNodes());
			}
			// remove duplicates
			return result.stream().distinct().collect(Collectors.toList());
		}
		case TRUE:
		case FALSE:
			return Collections.emptyList();
		case FIN:
		case FIN_NOT:
		case INF:
		case INF_NOT:
			return Collections.singletonList(this);
		}
		throw new UnsupportedOperationException("Unknown kind");
	}

	@Override
	public String getSignatureForState(int stateIndex) {
		List<AcceptanceELSummarized> leafNodes = getLeafNodes();

		String result = "";
		for (int i=0; i < leafNodes.size(); i++) {
			if (leafNodes.get(i).getStates().get(stateIndex)) {
				result += (result.isEmpty() ? "" : ",")+i;
			}
		}

		result = "{" + result + "}";
		return result;
	}

	@Override
	public String getSignatureForStateHOA(int stateIndex) {
		List<AcceptanceELSummarized> leafNodes = getLeafNodes();

		String result = "";
		for (int i=0; i < leafNodes.size(); i++) {
			if (leafNodes.get(i).getStates().get(stateIndex)) {
				result += (result.isEmpty() ? "" : " ")+i;
			}
		}

		if (!result.isEmpty())
			result = "{" + result + "}";
		return result;
	}

	@Override
	public String getSizeStatistics() {
		return "generic acceptance with " + countAcceptanceSets() + " acceptance sets";
	}

	@Override
	public AcceptanceType getType() {
		return AcceptanceType.GENERIC;
	}
	
	public AcceptanceELSummarized setToFalse(AcceptanceELSummarized fin)
	{
		switch(kind) {
		case TRUE:
		case FALSE:
		case INF:
		case INF_NOT:
		case FIN_NOT:
			return this;
		case FIN:
			return states.equals(fin.getStates()) ? new AcceptanceELSummarized(false) :
				this;
		case BIGAND:
			List<AcceptanceELSummarized> new_childs = new LinkedList<>();
			for(AcceptanceELSummarized child : childs) {
				AcceptanceELSummarized new_child = child.setToFalse(fin);
				if(new_child.getKind() == ElementType.FALSE) {
					return new AcceptanceELSummarized(false);
				}
				new_childs.add(new_child);
			}
			
			return new AcceptanceELSummarized(ElementType.BIGAND, new_childs);
		case BIGOR:
			new_childs = new LinkedList<>();
			for(AcceptanceELSummarized child : childs) {
				AcceptanceELSummarized new_child = child.setToFalse(fin);
				if(new_child.getKind() != ElementType.FALSE) {
					new_childs.add(new_child);
				}
			}
			
			if(new_childs.isEmpty()) {
				return new AcceptanceELSummarized(false);
			}
			if(new_childs.size() == 1) {
				return new_childs.get(0);
			}
			return new AcceptanceELSummarized(ElementType.BIGOR, new_childs);
			
		default:
			return null;
		}
	}

	@Override
	@Deprecated
	public String getTypeAbbreviated() {
		return getType().getNameAbbreviated();
	}

	@Override
	@Deprecated
	public String getTypeName() {
		return getType().getName();
	}

	@Override
	public AcceptanceELSummarized clone() {
		switch (kind) {
			case FIN:
			case FIN_NOT:
			case INF:
			case INF_NOT:
				return new AcceptanceELSummarized(kind, states);
			case BIGAND:
			case BIGOR:
				List<AcceptanceELSummarized> clonedChilds = new LinkedList<>();
				for(AcceptanceELSummarized child : childs) {
					clonedChilds.add(child.clone());
				}
				return new AcceptanceELSummarized(kind, clonedChilds);
			case FALSE:
				return new AcceptanceELSummarized(false);
			case TRUE:
				return new AcceptanceELSummarized(true);
		}
		throw new UnsupportedOperationException("Unsupported operator in generic acceptance condition");
	}

	/**
	 * Returns a new generic acceptance condition that corresponds to the conjunction
	 * of this and the other acceptance condition. Both conditions are <b>not</b>
	 * cloned; take care not to modify the conditions or clone beforehand.
	 * @param other the other generic acceptance condition
	 * @return new AcceptanceRabin, conjunction of this and other
	 */
	public AcceptanceELSummarized and(AcceptanceELSummarized other)
	{
		List<AcceptanceELSummarized> andChilds = new ArrayList<>(2);
		andChilds.add(this);
		andChilds.add(other);
		return new AcceptanceELSummarized(ElementType.BIGAND, andChilds);
	}

	/**
	 * Returns a new generic acceptance condition that corresponds to the disjunction
	 * of this and the other acceptance condition. Both conditions are <b>not</b>
	 * cloned; take care not to modify the conditions or clone beforehand.
	 * @param other the other generic acceptance condition
	 * @return new AcceptanceELSummarized, disjunction of this and other
	 */
	public AcceptanceELSummarized or(AcceptanceELSummarized other)
	{
		List<AcceptanceELSummarized> orChilds = new ArrayList<>(2);
		orChilds.add(this);
		orChilds.add(other);
		return new AcceptanceELSummarized(ElementType.BIGOR, orChilds);
	}

	/** Complement this acceptance condition, return as AcceptanceELSummarized. */
	public AcceptanceELSummarized complementToAcceptanceELSummarized()
	{
		switch (kind) {
		case TRUE: return new AcceptanceELSummarized(false);
		case FALSE:  return new AcceptanceELSummarized(true);

		case BIGAND:
			return new AcceptanceELSummarized(ElementType.BIGOR,
			                             childs.stream().map(child -> child.complementToAcceptanceELSummarized()).
			                             collect(Collectors.toList()));
		case BIGOR:
			return new AcceptanceELSummarized(ElementType.BIGAND,
			                             childs.stream().map(child -> child.complementToAcceptanceELSummarized()).
			                             collect(Collectors.toList()));
		case FIN:
			return new AcceptanceELSummarized(ElementType.INF, (BitSet) states.clone());
		case FIN_NOT:
			return new AcceptanceELSummarized(ElementType.INF_NOT, (BitSet) states.clone());
		case INF:
			return new AcceptanceELSummarized(ElementType.FIN, (BitSet) states.clone());
		case INF_NOT:
			return new AcceptanceELSummarized(ElementType.FIN_NOT, (BitSet) states.clone());
		default:
			throw new UnsupportedOperationException();
		}
	}

	@Override
	public AcceptanceOmega complement(int numStates, AcceptanceType... allowedAcceptance) throws PrismException
	{
		if (AcceptanceType.contains(allowedAcceptance, AcceptanceType.ELSUMMARIZED)) {
			return this.complementToAcceptanceELSummarized();
		}
		throw new PrismNotSupportedException("Can not complement " + getType() + " acceptance to required acceptance type");
	}

	@Override
	public void lift(LiftBitSet lifter) {
		switch(kind) {
			case TRUE:
			case FALSE:
				return;
			case INF:
			case INF_NOT:
			case FIN:
			case FIN_NOT:
				states = lifter.lift(states);
				return;
			case BIGAND:
			case BIGOR:
				for(AcceptanceELSummarized child : childs) {
					child.lift(lifter);
				}
				return;
		}
		throw new UnsupportedOperationException("Unsupported operator in generic acceptance condition");
	}

	/** Count the number of state sets in this acceptance condition */
	public int countAcceptanceSets() {
		switch(kind) {
			case FALSE:
			case TRUE:
				return 0;
			case INF:
			case FIN: 
			case INF_NOT:
			case FIN_NOT:
				return 1;
			case BIGOR:
			case BIGAND: 
				int result = 0;
				for(AcceptanceELSummarized child : childs) {
					result += child.countAcceptanceSets();
				}
				return result;
		}
		throw new UnsupportedOperationException("Unsupported operator in generic acceptance condition");
	}

	@Override
	public AcceptanceOmegaDD toAcceptanceDD(JDDVars ddRowVars) {
		return new AcceptanceELSummarizedDD(this, ddRowVars);
	}

	@Override
	public AcceptanceGeneric toAcceptanceGeneric()
	{
		switch(kind) {
		case TRUE:
			return new AcceptanceGeneric(true);
		case FALSE:
			return new AcceptanceGeneric(false);
		case FIN:
			return new AcceptanceGeneric(AcceptanceGeneric.ElementType.FIN, states);
		case FIN_NOT:
			return new AcceptanceGeneric(AcceptanceGeneric.ElementType.FIN_NOT, states);
		case INF:
			return new AcceptanceGeneric(AcceptanceGeneric.ElementType.INF, states);
		case INF_NOT:
			return new AcceptanceGeneric(AcceptanceGeneric.ElementType.INF_NOT, states);
		case BIGAND:
			if(childs.size() == 0) {
				return new AcceptanceGeneric(true);
			}

			if(childs.size() == 1) {
				return childs.get(0).toAcceptanceGeneric();
			}
			
			AcceptanceGeneric left = childs.get(0).toAcceptanceGeneric();
			AcceptanceELSummarized copied = clone();
			copied.getChilds().remove(0);
			AcceptanceGeneric right = copied.toAcceptanceGeneric();
			return new AcceptanceGeneric(AcceptanceGeneric.ElementType.AND, left, right);
		case BIGOR:
			if(childs.size() == 0) {
				return new AcceptanceGeneric(true);
			}

			if(childs.size() == 1) {
				return childs.get(0).toAcceptanceGeneric();
			}
			
			left = childs.get(0).toAcceptanceGeneric();
			copied = clone();
			copied.getChilds().remove(0);
			right = copied.toAcceptanceGeneric();
			return new AcceptanceGeneric(AcceptanceGeneric.ElementType.OR, left, right);
		default:
			break;
		}
		return null;
	}

	@Override
	public String toString() {
		switch(kind) {
			case TRUE:
				return "true";
			case FALSE:
				return "false";
			case BIGAND:
				return String.join(" & ", childs.stream().
													map(child -> child.toString()).
													collect(Collectors.toList()));
			case BIGOR:
				return String.join(" | ", childs.stream().
													map(child -> child.toString()).
													collect(Collectors.toList()));
			case INF:
				return "Inf(" + states.toString() + ")";
			case FIN:
				return "Fin(" + states.toString() + ")";
			case INF_NOT:
				return "Inf(!" + states.toString() + ")";
			case FIN_NOT:
				return "Fin(!" + states.toString() + ")";
			default:
				return null;
		}
	}

	@Override
	public void outputHOAHeader(PrintStream out)
	{
		List<AcceptanceELSummarized> leafNodes = getLeafNodes();
		out.print("Acceptance: "+leafNodes.size()+" ");
		outputHOAFormula(out, 0);
		out.println();
	}

	private int outputHOAFormula(PrintStream out, int nextSetIndex)
	{
		switch (kind) {
		case BIGAND:
			boolean first = true;
			for(AcceptanceELSummarized child: childs) {
				if(first) {
					first = false;
				} else {
					out.print("&");
				}
				out.print("(");
				nextSetIndex = child.outputHOAFormula(out, nextSetIndex);
				out.print(")");
			}
			return nextSetIndex;
		case BIGOR:
			first = true;
			for(AcceptanceELSummarized child: childs) {
				if(first) {
					first = false;
				} else {
					out.print("|");
				}
				out.print("(");
				nextSetIndex = child.outputHOAFormula(out, nextSetIndex);
				out.print(")");
			}
			return nextSetIndex;
		case TRUE:
			out.print("t");
			return nextSetIndex;
		case FALSE:
			out.print("f");
			return nextSetIndex;
		case FIN:
			out.print("Fin("+nextSetIndex+")");
			return nextSetIndex+1;
		case FIN_NOT:
			out.print("Fin(!"+nextSetIndex+")");
			return nextSetIndex+1;
		case INF:
			out.print("Inf("+nextSetIndex+")");
			return nextSetIndex+1;
		case INF_NOT:
			out.print("Inf(!"+nextSetIndex+")");
			return nextSetIndex+1;
		}
		throw new UnsupportedOperationException("Unknown kind");
	}
	
	public AcceptanceELSummarized simplify() throws PrismNotSupportedException
	{
		switch(kind) {
		case TRUE:
		case FALSE:
			return this;
		case FIN:
			return (states.isEmpty()) ? new AcceptanceELSummarized(true) : this;
		case INF:
			return (states.isEmpty()) ? new AcceptanceELSummarized(false) : this;
		case BIGAND:
			// Collect every fin
			List<AcceptanceELSummarized> newChilds = new ArrayList<>();
			BitSet finStates = new BitSet();
			for(AcceptanceELSummarized child : childs) {
				if(child.getKind() == ElementType.FIN) {
					finStates.or(child.getStates());
				} else {
					AcceptanceELSummarized newChild = child.simplify();
					if(newChild.getKind() == ElementType.FALSE) {
						return new AcceptanceELSummarized(false);
					}

					if(newChild.getKind() != ElementType.TRUE) {
						newChilds.add(newChild);
					}
				}
			}
			if(newChilds.isEmpty() && finStates.isEmpty()) {
				return new AcceptanceELSummarized(true);
			}
			// Remove fin states from the remainder formula
			List<AcceptanceELSummarized> resultChilds;
			if(finStates.isEmpty()) {
				return new AcceptanceELSummarized(ElementType.BIGAND,
						newChilds);
			} else {
				resultChilds = newChilds.stream().
						map(child -> child.intersect(finStates, true)).
						collect(Collectors.toList());

				if(resultChilds.stream().anyMatch(child -> child.getKind() == ElementType.FALSE)) {
					return new AcceptanceELSummarized(false);
				}
				resultChilds.add(0, new AcceptanceELSummarized(ElementType.FIN, finStates));
				return new AcceptanceELSummarized(ElementType.BIGAND,
						resultChilds);
			} 
		case BIGOR:
			// Collect every Inf
			newChilds = new ArrayList<>();
			BitSet infStates = new BitSet();
			for(AcceptanceELSummarized child : childs) {
				if(child.getKind() == ElementType.INF) {
					infStates.or(child.getStates());
				} else {
					AcceptanceELSummarized newChild = child.simplify();
					if(newChild.getKind() == ElementType.TRUE) {
						return new AcceptanceELSummarized(true);
					}
					if(newChild.getKind() != ElementType.FALSE) {
						newChilds.add(newChild);
					}
				}
			}
			if(newChilds.isEmpty() && infStates.isEmpty()) {
				return new AcceptanceELSummarized(false);
			}
			// Remove fin states from the remainder formula
			if(infStates.isEmpty()) {
				return new AcceptanceELSummarized(ElementType.BIGOR,
						newChilds);
			} else {
				if(newChilds.stream().anyMatch(child -> child.getKind() == ElementType.TRUE)) {
					return new AcceptanceELSummarized(true);
				}
				newChilds.add(0, new AcceptanceELSummarized(ElementType.INF, infStates));
				return new AcceptanceELSummarized(ElementType.BIGOR,
						newChilds);
			}
		case FIN_NOT:
		case INF_NOT:
			throw new PrismNotSupportedException("Simplification for FinNot and InfNot are not supported");
		}
		return null;
	}
	
	public String structureToString() {
		switch(kind) {
			case TRUE:
				return "true";
			case FALSE:
				return "false";
			case BIGAND:
				return "(" + 
						String.join(" & ", childs.stream().
													map(child -> child.structureToString()).
													collect(Collectors.toList()))
						+ ")";
			case BIGOR:
				return "(" +
						String.join(" | ", childs.stream().
													map(child -> child.structureToString()).
													collect(Collectors.toList()))
						+ ")";
			case INF:
				return "Inf(•)";
			case FIN:
				return "Fin(•)";
			case INF_NOT:
				return "Inf(!•)";
			case FIN_NOT:
				return "Fin(!•)";
			default:
				return null;
		}
	}
	
	public Set<AcceptanceELSummarized> getLeafs() {
		switch(kind) {
			case TRUE:
			case FALSE:
				return new HashSet<>();
			case BIGAND:
			case BIGOR:
				Set<AcceptanceELSummarized> result = new HashSet<>();
				for(AcceptanceELSummarized child : childs) {
					result.addAll(child.getLeafs());
				}
				return result;
			case INF:
			case FIN:
			case INF_NOT:
			case FIN_NOT:
				result = new HashSet<>();
				result.add(this);
				return result;
			default:
				return null;
		}
		
	}


}
