//	* Dave Parker <d.a.parker@cs.bham.ac.uk> (University of Birmingham/Oxford)
//  * David Müller <david.mueller2@tu-dresden.de> (TU Dresden)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package explicit.satsolving;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import acceptance.AcceptanceGenericDD;
import acceptance.AcceptanceOmegaDD;
import acceptance.AcceptanceType;
import jdd.JDD;
import jdd.JDDNode;
import jdd.JDDVars;
import prism.PrismNotSupportedException;

/**
 * A generic acceptance condition (based on BitSet state sets).
 * This is an AST of a boolean formula (conjunction and disjunction) over
 * atoms of the form Inf(states), Inf(!states), Fin(states) and Fin(!states).
 * <br>
 * Semantics:
 *  Inf(states)  <=> G F states
 *  Inf(!states) <=> G F !states
 *  Fin(states)  <=> F G !states
 *  Fin(!states) <=> F G states
 */
public class AcceptanceELSummarizedDD implements AcceptanceOmegaDD {
	
	/** The types of elements in the AST */
	public enum ElementType {
		FALSE,
		TRUE, 

		BIGOR,
		BIGAND,

		INF,
		FIN,
		INF_NOT,
		FIN_NOT;
	}
	
	/** The type of this node in the AST */
	private ElementType kind;

	/** The childs (if it exist) */ 
	private List<AcceptanceELSummarizedDD> childs = new LinkedList<>();

	/** The set of states (if this is one of INF, FIN, INF_NOT, FIN_NOT) */
	private JDDNode states = null;

	/**
	 * Constructor for TRUE or FALSE
	 * @param value true or false?
	 */
	public AcceptanceELSummarizedDD(boolean value) {
		kind = value ? ElementType.TRUE : ElementType.FALSE;
	}
	
	/** 
	 * Constructor for an INF, FIN, INF_NOT or FIN_NOT element. 
	 */
	public AcceptanceELSummarizedDD(ElementType kind, JDDNode states) {
		this.kind = kind;

		this.states = states;
	}

	/**
	 * Constructor for a binary operator (AND/OR).
	 * @param kind
	 * @param left
	 * @param right
	 */
	public AcceptanceELSummarizedDD(ElementType kind, List<AcceptanceELSummarizedDD> childs) {
		assert(kind == ElementType.BIGAND || kind == ElementType.BIGOR);
		this.kind = kind;
		
		for(AcceptanceELSummarizedDD child: childs) {
			if(child.getKind() == kind) {
				childs.addAll(child.getChilds());
			} else {
				this.childs.add(child);
			}
		}
	}

	public AcceptanceELSummarizedDD(AcceptanceELSummarized acceptanceELSummarized, JDDVars ddRowVars) {
		switch(acceptanceELSummarized.getKind()) {
		case BIGAND:
			break;
		case BIGOR:
			break;
		case FALSE:
			break;
		case FIN:
			break;
		case FIN_NOT:
			break;
		case INF:
			break;
		case INF_NOT:
			break;
		case TRUE:
			break;
		default:
			break;
		}
	}

	public List<AcceptanceELSummarizedDD> getChilds() {
		return childs;
	}

	/** Get the ElementType of this AST element */
	public ElementType getKind() {
		return kind;
	}

	/** Get the state set of this element (if kind is one of INF, FIN, INF_NOT, FIN_NOT).
	 */
	public JDDNode getStates() {
		return states;
	}

	@Override
	public boolean isBSCCAccepting(JDDNode bscc) {
		switch(kind) {
		case TRUE: return true;
		case FALSE: return false;
		case BIGAND:
			return childs.stream().allMatch(child -> child.isBSCCAccepting(bscc));
		case BIGOR:
			return childs.stream().anyMatch(child -> child.isBSCCAccepting(bscc));
		case INF:
			// bscc |= G F states?
			// there exists a state in bscc and states
			return JDD.AreIntersecting(states, bscc);
		case INF_NOT:
			// bscc_state |= G F !states?
			// the BSCC intersects Not(states)
			JDD.Ref(states);
			return JDD.AreIntersecting(JDD.Not(states), bscc);
		case FIN:
			// bscc |= F G !states?
			// the BSCC consists only of !states
			return !JDD.AreIntersecting(states, bscc);
		case FIN_NOT:
			// bscc |= F G states?
			// the BSCC consists entirely of states
			JDD.Ref(states);
			return !JDD.AreIntersecting(JDD.Not(states), bscc);
		}
		return false;
	}

	/** Get a list of all the (non-true/false) leaf nodes in this acceptance condition */
	public List<AcceptanceELSummarizedDD> getLeafNodes()
	{
		switch (getKind()) {
		case BIGAND:
		case BIGOR: {
			List<AcceptanceELSummarizedDD> result = new ArrayList<>();
			for(AcceptanceELSummarizedDD child: childs) {
				result.addAll(child.getLeafNodes());
			}
			// remove duplicates
			return result.stream().distinct().collect(Collectors.toList());
		}
		case TRUE:
		case FALSE:
			return Collections.emptyList();
		case FIN:
		case FIN_NOT:
		case INF:
		case INF_NOT:
			return Collections.singletonList(this);
		}
		throw new UnsupportedOperationException("Unknown kind");
	}

	@Override
	public String getSizeStatistics() {
		return "generic acceptance with " + countAcceptanceSets() + " acceptance sets";
	}

	@Override
	public AcceptanceType getType() {
		return AcceptanceType.GENERIC;
	}

	@Override
	@Deprecated
	public String getTypeAbbreviated() {
		return getType().getNameAbbreviated();
	}

	@Override
	@Deprecated
	public String getTypeName() {
		return getType().getName();
	}

	@Override
	public AcceptanceELSummarizedDD clone() {
		switch (kind) {
			case FIN:
			case FIN_NOT:
			case INF:
			case INF_NOT:
				return new AcceptanceELSummarizedDD(kind, states);
			case BIGAND:
			case BIGOR:
				List<AcceptanceELSummarizedDD> clonedChilds = new LinkedList<>();
				for(AcceptanceELSummarizedDD child : childs) {
					clonedChilds.add(child.clone());
				}
				return new AcceptanceELSummarizedDD(kind, clonedChilds);
			case FALSE:
				return new AcceptanceELSummarizedDD(false);
			case TRUE:
				return new AcceptanceELSummarizedDD(true);
		}
		throw new UnsupportedOperationException("Unsupported operator in generic acceptance condition");
	}

	/**
	 * Returns a new generic acceptance condition that corresponds to the conjunction
	 * of this and the other acceptance condition. Both conditions are <b>not</b>
	 * cloned; take care not to modify the conditions or clone beforehand.
	 * @param other the other generic acceptance condition
	 * @return new AcceptanceRabin, conjunction of this and other
	 */
	public AcceptanceELSummarizedDD and(AcceptanceELSummarizedDD other)
	{
		List<AcceptanceELSummarizedDD> andChilds = new ArrayList<>(2);
		andChilds.add(this);
		andChilds.add(other);
		return new AcceptanceELSummarizedDD(ElementType.BIGAND, andChilds);
	}

	/**
	 * Returns a new generic acceptance condition that corresponds to the disjunction
	 * of this and the other acceptance condition. Both conditions are <b>not</b>
	 * cloned; take care not to modify the conditions or clone beforehand.
	 * @param other the other generic acceptance condition
	 * @return new AcceptanceELSummarizedDD, disjunction of this and other
	 */
	public AcceptanceELSummarizedDD or(AcceptanceELSummarizedDD other)
	{
		List<AcceptanceELSummarizedDD> orChilds = new ArrayList<>(2);
		orChilds.add(this);
		orChilds.add(other);
		return new AcceptanceELSummarizedDD(ElementType.BIGOR, orChilds);
	}

	/** Complement this acceptance condition, return as AcceptanceELSummarizedDD. */
	public AcceptanceELSummarizedDD complementToAcceptanceELSummarizedDD()
	{
		switch (kind) {
		case TRUE: return new AcceptanceELSummarizedDD(false);
		case FALSE:  return new AcceptanceELSummarizedDD(true);

		case BIGAND:
			return new AcceptanceELSummarizedDD(ElementType.BIGOR,
			                             childs.stream().map(child -> child.complementToAcceptanceELSummarizedDD()).
			                             collect(Collectors.toList()));
		case BIGOR:
			return new AcceptanceELSummarizedDD(ElementType.BIGAND,
			                             childs.stream().map(child -> child.complementToAcceptanceELSummarizedDD()).
			                             collect(Collectors.toList()));
		case FIN:
			return new AcceptanceELSummarizedDD(ElementType.INF, (JDDNode) states);
		case FIN_NOT:
			return new AcceptanceELSummarizedDD(ElementType.INF_NOT, (JDDNode) states);
		case INF:
			return new AcceptanceELSummarizedDD(ElementType.FIN, (JDDNode) states);
		case INF_NOT:
			return new AcceptanceELSummarizedDD(ElementType.FIN_NOT, (JDDNode) states);
		default:
			throw new UnsupportedOperationException();
		}
	}

	/** Count the number of state sets in this acceptance condition */
	public int countAcceptanceSets() {
		switch(kind) {
			case FALSE:
			case TRUE:
				return 0;
			case INF:
			case FIN: 
			case INF_NOT:
			case FIN_NOT:
				return 1;
			case BIGOR:
			case BIGAND: 
				int result = 0;
				for(AcceptanceELSummarizedDD child : childs) {
					result += child.countAcceptanceSets();
				}
				return result;
		}
		throw new UnsupportedOperationException("Unsupported operator in generic acceptance condition");
	}

	@Override
	public String toString() {
		switch(kind) {
			case TRUE:
				return "true";
			case FALSE:
				return "false";
			case BIGAND:
				return String.join(" & ", childs.stream().
													map(child -> child.toString()).
													collect(Collectors.toList()));
			case BIGOR:
				return String.join(" | ", childs.stream().
													map(child -> child.toString()).
													collect(Collectors.toList()));
			case INF:
				return "Inf(" + states.toString() + ")";
			case FIN:
				return "Fin(" + states.toString() + ")";
			case INF_NOT:
				return "Inf(!" + states.toString() + ")";
			case FIN_NOT:
				return "Fin(!" + states.toString() + ")";
			default:
				return null;
		}
	}

	@Override
	public void intersect(JDDNode restrict) {
		switch(kind) {
		case FALSE:
		case TRUE:
			break;
		case BIGAND:
		case BIGOR:
			for(AcceptanceELSummarizedDD child : childs) {
				child.intersect(restrict);
			}
		case INF:
		case INF_NOT:
		case FIN:
		case FIN_NOT:
			states = JDD.And(states, restrict.copy());
		default:
			break;
		}
	}

	@Override
	public AcceptanceOmegaDD complement(AcceptanceType... allowedAcceptance) throws PrismNotSupportedException {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public AcceptanceGenericDD toAcceptanceGeneric() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public void clear() {
		// TODO Auto-generated method stub
		
	}

}
