package explicit.satsolving;

import java.util.BitSet;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import java.util.stream.Collectors;

class Clause extends LinkedList<Literal>
{
		
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	public enum Assignment {
			TRUE,
			FALSE,
			UNASSIGNED
	}
	
	/**
	 * signs, whether the clause is false, true or unassigned
	 */
	protected Assignment value = Assignment.UNASSIGNED;
	
	/**
	 * Checks, whether this is equivalent to true after simple transformations 
	 * @return true, if this is equivalent to true
	 */
	public boolean isTrue() {
		recalculateAssignment();
		return value == Assignment.TRUE;
	}
	
	/**
	 * Sets this to true
	 */
	public void setTrue() {
		value = Assignment.TRUE;
	}
	
	
	/**
	 * Performs an intersection between the set of states in every FIN and INF literal and mec
	 * @param mec the set of states the literals should be intersected with
	 */
	public void cutWithMEC(BitSet mec)
	{
		Iterator<Literal> it = iterator();
		while(it.hasNext()) {
			Literal literal = it.next();
			//Go through every literal, intersect, if it is of the form INF or FIN
			if(literal.getKind() == Literal.Kind.INF || literal.getKind() == Literal.Kind.FIN) {
				literal.getStates().and(mec);
				if(literal.getStates().isEmpty()) {
					if(literal.getKind() == Literal.Kind.INF) {
						literal.setFalse();
					}
					if(literal.getKind() == Literal.Kind.FIN) {
						literal.setTrue();
						setTrue();
					}
				}
			}
		}
		simplify();
	}
	
	/**
	 * If two Infs-Literal INF(P1) and INF(P2) appears in a clause, remove both and
	 * add INF(P1 \cup P2)
	 */
	public void unifyAllInfs()
	{
		BitSet inf = new BitSet();
		Iterator<Literal> it = iterator();
		while(it.hasNext()) {
			Literal ap = it.next();

			if(ap.getKind() == Literal.Kind.INF) {
				//Store the states in inf and
				//remove the ap
				inf.or(ap.getStates());
				it.remove();
			}
		}
		
		if(!inf.isEmpty()) {
			add(new Literal(inf, true));
		}
	}
	
	/**
	 * Performs simple transformation preserving equivalence
	 */
	public void simplify()
	{
		if(value != Assignment.TRUE && value != Assignment.FALSE) {
			//Search for empty Fin
			Iterator<Literal> it = iterator();
			while(it.hasNext()) {
				Literal ap = it.next();
				if(ap.getKind() == Literal.Kind.FIN && ap.getStates().isEmpty()) {
					ap.setTrue();
					value = Assignment.TRUE;
					return;
				}
			}
		
			//TODO: Set clause to true, if the are aps of the form p and not p (and analogously Fin and Inf)
		
			//TODO: Remove Fin(S'), if there is a Fin(S) with S \subseteq S'
		
		}
		
	}
	
	/**
	 * Returns a textual representation of this clause
	 */
	public String toString() 
	{
		if(value == Assignment.TRUE) {
			return "true";
		}
		
		if(value == Assignment.FALSE) {
			return "false";
		}
		
		String result = "";
		Iterator<Literal> it = iterator();
		if(!it.hasNext()) {
			return "false";
		}
		Literal literal = it.next();;
		while(literal.isFalse() && it.hasNext()) {
			literal = it.next();
		}
		if(literal.isFalse() && !it.hasNext()) {
			return "false";
		}
		result = literal.toString();
		while(it.hasNext()) {
			literal = it.next();
			if(!literal.isTrue()) {
				result += " | " + literal.toString();
			}
		}
		return result;
	}

	/**
	 * Performs simple transformations preserving equivalence, and
	 * subsequently decides, whether this is false, true or unassigned
	 */
	public void recalculateAssignment() 
	{
		simplify();
		//An empty clause is false
		if(size() == 0) {
			value = Assignment.FALSE;
			return;
		}
		//A true literal implies that the clause is true
		for(Literal literal : this) {
			if(literal.isTrue()) {
				value = Assignment.TRUE;
				return;
			}
		}
		//Search for unassigned literals
		Set<Literal> unassignedList = stream().filter(literal -> literal.isUnassigned()).collect(Collectors.toSet());
		if(unassignedList.isEmpty()) {
			//If there are no unassigned literals, and the clause is not true, it is false
			value = Assignment.FALSE;
		} else {
			//If there is an unassigned literal (and no true literals)
			//then a clause is unassigned
			value = Assignment.UNASSIGNED;
		}
	}

	/**
	 * Checks, whether this is equivalent to false after simple transformations 
	 * @return true, if this is equivalent to false
	 */
	public boolean isFalse() {
		recalculateAssignment();
		return value == Assignment.FALSE;
	}
	
	/**
	 * Checks, whether this is a unit clause 
	 * Note that, (INF(P) | FIN(Q)), (INF(P) | s), (INF(P) | !s) are not unit clauses
	 * @return the literal in the unit, if this is a unit clause, null otherwise
	 */
	public Literal isUnitClause()
	{
		if(value != Assignment.UNASSIGNED) {
			return null;
		}
		
		Set<Literal> unassignedList = stream().filter(literal -> literal.isUnassigned()).collect(Collectors.toSet());
		if(unassignedList.size() != 1) {
			return null;
		}
		Iterator<Literal> it = unassignedList.iterator();
		it.hasNext();
		Literal unitLit = it.next();
		if(unitLit.getKind() != Literal.Kind.INF) {
			return unitLit.clone();
		}
		return null;
	}
	
	public void backtrack(Literal literal, boolean value, BitSet mec)
	{
		for(Literal clauseLit : this) {
			clauseLit.backtrack(literal, value, mec);
			if(clauseLit.isTrue()) {
				this.value = Assignment.TRUE;
			}
			if(this.value == Assignment.FALSE && clauseLit.isUnassigned()) {
				this.value = Assignment.UNASSIGNED;
			}
		}
	}
	
	/**
	 * returns a deep copy of this
	 */
	public Clause clone()
	{
		Clause clause = new Clause();
		for(Literal literal : this) {
			clause.add(literal.clone());
		}
		return clause;
	}
	
	/**
	 * Checks, whether this is unassigned
	 * @return true, if this is unassigned
	 */
	public boolean isUnassigned()
	{
		return value==Assignment.UNASSIGNED;
	}

	/**
	 * Remove all false Literals in a clause
	 */
	public void removeFalse()
	{
		for(Iterator<Literal> it = iterator(); it.hasNext(); ) {
			Literal lit = it.next();
			if(lit.isFalse()) {
				it.remove();
			}
		}
	}
	
	public void removeDuplicates()
	{
		Set<Literal> literals = new HashSet<Literal>();
		for(Iterator<Literal> it = iterator(); it.hasNext();) {
			Literal lit = it.next();
			if(literals.contains(lit)) {
				it.remove();
			} else {
				literals.add(lit);
			}
		}
	}
}
