package explicit.satsolving;

import java.util.BitSet;
import java.util.NoSuchElementException;
import java.util.Stack;

import acceptance.AcceptanceGeneric;
import prism.PrismException;

public class Literal implements Comparable<Literal> {
	public enum Assignment {
			TRUE,
			FALSE,
			UNASSIGNED
	}
	
	public enum Kind {
		INF,
		FIN,
		SLACKVAR,
		NEGSLACKVAR
	}

	private int index;
	private BitSet states;
	private Assignment value = Assignment.UNASSIGNED;
	private Kind kind;
	private Stack<BitSet> originalStates = new Stack<>();
	
	protected Clause refClause;
	protected int decisionLevel;
	
	/**
	 * Returns a deep copy of this
	 */
	public Literal clone()
	{
		Literal literal = null;
		switch(kind) {
		case INF:
			literal = new Literal(getStates(), true);
			break;
		case FIN:
			literal = new Literal(getStates(), false);
			break;
		case SLACKVAR:
			literal = new Literal(index, false);
			break;
		case NEGSLACKVAR:
			literal = new Literal(index, true);
			break;
		default:
				
		}
		literal.decisionLevel = decisionLevel;
		literal.value = value;
		literal.refClause = refClause;
		return literal;
	}
	
	public Literal(int index, boolean negated) 
	{
		this.index = index;
		kind = negated ? Kind.NEGSLACKVAR : Kind.SLACKVAR;
	}
	
	public Literal(BitSet states, boolean inf)
	{
		this.states = (BitSet)(states.clone());
		kind = inf ? Kind.INF : Kind.FIN;
	}
	
	public Literal(AcceptanceGeneric acc) throws PrismException
	{
		switch(acc.getKind()) {
		case FIN:
			kind = Kind.FIN;
			states = (BitSet)acc.getStates().clone();
			break;
		case INF:
			kind = Kind.INF;
			states = (BitSet)acc.getStates().clone();
			break;
		default:
			throw new PrismException("Not implemented");
		}
	}

	/**
	 * Checks, whether this is contained in the saved clause (ref)
	 * @return true, if ref contains this
	 */
	public boolean isContainedInRefClause()
	{
		//For INF and FIN, a simple element test suffices
		if(kind == Kind.INF || kind == Kind.FIN) {
			return refClause.contains(this);
		}
		
		//Negate Variable
		return refClause.contains(this) ||
				refClause.contains(complement());
	}

	public boolean backtrackFinValue(Literal literal, boolean value, BitSet mec)
	{
		if(!isUnassigned()) {
			switch(getKind()) {
				case FIN:
					if(value) {
						BitSet lastStates = originalStates.pop();
						states = lastStates;
						if(Utils.isSubsetEq(getStates(), literal.getStates())) {
							setUnassigned();
							return true;
						} else {
							return false;
						}
					} else {
						if(Utils.isSubsetEq(literal.getStates(), originalStates.peek())) {
							states = originalStates.pop();
							setUnassigned();
							return true;
						} else {
							return false;
						}
						
					}
				case INF:
					if(value) {
						states = originalStates.pop();
						if(Utils.isSubsetEq(getStates(), literal.getStates())) {
							setUnassigned();
							return true;
						} else {
							return false;
						}
					} else {
						if(Utils.isSubsetEq(literal.getStates(), originalStates.peek())) {
							states = originalStates.pop();
							setUnassigned();
							return true;
						} else {
							return false;
						}
						
					}
				default:
					return false;
			}
		} else {
			return false;
		}
	}
	
	public boolean backtrackSlackVarValue(Literal literal, boolean value, BitSet mec)
	{
		switch(getKind()) {
			case SLACKVAR:
				if(getIndex() == literal.getIndex()) {
					if(value) {
						setUnassigned();
						return true;
					} else {
						setUnassigned();
						return true;
					}
				}
				return false;
			case NEGSLACKVAR:
				if(getIndex() == literal.getIndex()) {
					if(value) {
						setUnassigned();
						return true;
					} else {
						setUnassigned();
						return true;
					}
				}
				return false;
			default:
				return false;
		}
	}
	
	public boolean backtrackNegSlackVarValue(Literal literal, boolean value, BitSet mec)
	{
		return backtrackSlackVarValue(literal.complement(), !value, mec);
	}
	
	public boolean backtrack(Literal literal, boolean value, BitSet mec)
	{
		switch(literal.getKind()) {
		case SLACKVAR: return backtrackSlackVarValue(literal, value, mec);
		case NEGSLACKVAR: return backtrackNegSlackVarValue(literal, value, mec);
		case FIN: return backtrackFinValue(literal, value, mec);
		default:
			return false;
		}
	}
	
	
	public boolean equals(Object o)
	{
		if(!(o instanceof Literal)) {
			return false;
		}
		
		Literal other = (Literal)o;
		if(other.getKind() != getKind()) {
			return false;
		}
		
		switch(getKind()) {
		case INF:
		case FIN:
			return getStates().equals(other.getStates());
		case SLACKVAR:
		case NEGSLACKVAR:
			return getIndex() == other.getIndex();
		}
		return false;
	}
	
	public int hashCode() 
	{
		switch(getKind()) {
		case SLACKVAR:
			return getIndex();
		case NEGSLACKVAR:
			return ~getIndex();
		case INF:
			return getStates().hashCode();
		case FIN:
			return ~getStates().hashCode();
		}
		
		return 0;
	}
	
	/**
	 * signals the value of the literal
	 * @return true, if this is false
	 */
	public boolean isFalse()
	{
		return value == Assignment.FALSE;
	}
	
	/**
	 * sets this to false
	 */
	public void setFalse()
	{
		value = Assignment.FALSE;
	}

	/**
	 * signals the value of the literal
	 * @return true, if this is true
	 */
	public boolean isTrue()
	{
		return value == Assignment.TRUE;
	}
	
	/**
	 * sets this to true
	 */
	public void setTrue()
	{
		value = Assignment.TRUE;
	}
	
	/**
	 * signals the value of the literal
	 * @return true, if this is unassigned
	 */
	public boolean isUnassigned()
	{
		return value == Assignment.UNASSIGNED;
	}
	
	/**
	 * sets this to unassigned
	 */
	public void setUnassigned()
	{
		value = Assignment.UNASSIGNED;
	}

	public Clause getRefClause() {
		return refClause;
	}
	public void setRefClause(Clause refClause) {
		this.refClause = refClause;
	}
	public int getDecisionLevel() {
		return decisionLevel;
	}

	public void setDecisionLevel(int decisionLevel) {
		this.decisionLevel = decisionLevel;
	}
	public int getIndex() {
		return index;
	}
	
	public Kind getKind() 
	{
		return kind;
	}
	
	public BitSet getStates()
	{
		return states;
	}
	
	public boolean assignSlackVarValue(Literal literal, boolean value, BitSet mec)
	{
		switch(getKind()) {
			case SLACKVAR:
				if(getIndex() == literal.getIndex()) {
					if(value) {
						setTrue();
						return true;
					} else {
						setFalse();
						return true;
					}
				}
				return false;
			case NEGSLACKVAR:
				if(getIndex() == literal.getIndex()) {
					if(value) {
						setFalse();
						return true;
					} else {
						setTrue();
						return true;
					}
				}
				return false;
			default:
				return false;
		}
	}
	
	public boolean assignNegSlackVarValue(Literal literal, boolean value, BitSet mec)
	{
		return assignSlackVarValue(literal.complement(), !value, mec);
	}
	
	private Literal complement() {
		switch(getKind()) {
		case SLACKVAR:
			return new Literal(getIndex(), true);
		case NEGSLACKVAR:
			return new Literal(getIndex(), false);
		case INF:
			return new Literal(getStates(), false);
		case FIN:
			return new Literal(getStates(), true);
		default:
			return null;
		}
	}

	public boolean assignFinValue(Literal literal, boolean value, BitSet mec)
	{
		switch(getKind()) {
			case FIN:
				if(value) {
					originalStates.push((BitSet)getStates().clone());
					if(Utils.isSubsetEq(getStates(), literal.getStates())) {
						setTrue();
						return true;
					} else {
						states.andNot(literal.getStates());
						return false;
					}
				} else {
					if(Utils.isSubsetEq(literal.getStates(), getStates())) {
						originalStates.push((BitSet)getStates().clone());
						setFalse();
						return true;
					} else {
						return false;
					}
					
				}
			case INF:
				if(value) {
					originalStates.push((BitSet)getStates().clone());
					if(Utils.isSubsetEq(getStates(), literal.getStates())) {
						setFalse();
						return true;
					} else {
						states.andNot(literal.getStates());
						return false;
					}
				} else {
					if(Utils.isSubsetEq(literal.getStates(), getStates())) {
						originalStates.push((BitSet)getStates().clone());
						setTrue();
						return true;
					} else {
						return false;
					}
					
				}
			default:
				return false;
		}
	}
	
	/**
	 * Assigns the value to this, if it literal agrees with this, i.e.,
	 * they have corresponding kinds
	 * (e.g. SLACKVAR and SLACKVAR or FIN and INF with corresponding states sets)
	 * 
	 * @param literal the literal to which this should be compared
	 * @param value signs, whether true or false should be assigned
	 * @return true, if an assignment happened, false otherwise
	 * @throws PrismException 
	 */
	public boolean assignValue(Literal literal, boolean value, BitSet mec) 
	{
		switch(literal.getKind()) {
		case FIN:
			return assignFinValue(literal, value, mec);
		case SLACKVAR:
			return assignSlackVarValue(literal, value, mec);
		case NEGSLACKVAR:
			return assignNegSlackVarValue(literal, value, mec);
		default:
			return false;
		}
	}
	
	public String toString()
	{
		switch(kind) {
		case SLACKVAR: return "s" + index;
		case NEGSLACKVAR: return "!s" + index;
		case INF: return "Inf(" + getStates() + ")";
		case FIN: return "Fin(" + getStates() + ")";
		default:
			throw new NoSuchElementException();
		}
	}

	@Override
	public int compareTo(Literal other) 
	{
		if(getKind() == Kind.SLACKVAR) {
			if(other.getKind() == Kind.SLACKVAR) {
				return Integer.valueOf(getIndex()).compareTo(Integer.valueOf(other.getIndex()));
			}
			if(other.getKind() == Kind.NEGSLACKVAR) {
				return getIndex() <= other.getIndex() ? -1 : 1; 
			}
			return -1;
		}

		if(getKind() == Kind.NEGSLACKVAR) {
			if(other.getKind() == Kind.NEGSLACKVAR) {
				return Integer.valueOf(getIndex()).compareTo(Integer.valueOf(other.getIndex()));
			}
			if(other.getKind() == Kind.SLACKVAR) {
				return other.getIndex() <= getIndex() ? 1 : -1; 
			}
			return -1;
		}

		if(getKind() == Kind.FIN) {
			if(other.getKind() == Kind.SLACKVAR ||
					other.getKind() == Kind.NEGSLACKVAR) {
				return 1;
			}
			
			if(other.getKind() == Kind.FIN) {
				if(getStates().size() > other.getStates().size()) {
					return -1;
				}
				if(getStates().size() < other.getStates().size()) {
					return 1;
				}

				// Just do some stuff to get a linear order ;-)
				
				for(int i = 0; i < getStates().size(); i++) {
					if(getStates().get(i) && !other.getStates().get(i)) {
						return -1;
					}
					if(!getStates().get(i) && other.getStates().get(i)) {
						return 1;
					}
				}
				
				// both literals equals
				return 0;
			}
			// Other literal is a INF literal
			return -1;
		}

		if(getKind() == Kind.INF) {
			if(other.getKind() == Kind.INF) {
				if(getStates().size() < other.getStates().size()) {
					return -1;
				}
				if(getStates().size() > other.getStates().size()) {
					return 1;
				}

				// Just do some stuff to get a linear order ;-)
				
				for(int i = 0; i < getStates().size(); i++) {
					if(getStates().get(i) && !other.getStates().get(i)) {
						return 1;
					}
					if(!getStates().get(i) && other.getStates().get(i)) {
						return -1;
					}
				}
				
				// both literals equals
				return 0;
			}
			// Other literal is a INF literal
			return 1;
		}
		return 0;
	}
	
	public void setStates(BitSet states)
	{
		this.states = states;
	}
}
