package symbolic.satsolving;

import java.util.List;
import java.util.NoSuchElementException;
import java.util.Stack;

import acceptance.AcceptanceGenericDD;
import jdd.JDD;
import jdd.JDDNode;
import prism.ECComputer;
import prism.NondetModel;
import prism.PrismComponent;
import prism.PrismException;

public abstract class AbstractLiteral implements Comparable<AbstractLiteral> {
	public enum Assignment {
			TRUE,
			FALSE,
			UNASSIGNED
	}
	
	public enum Kind {
		INF,
		FIN,
		SLACKVAR,
		NEGSLACKVAR
	}

	protected int index;
	protected JDDNode states;
	protected Assignment value = Assignment.UNASSIGNED;
	protected Kind kind;
	protected Stack<JDDNode> originalStates = new Stack<>();
	protected int numVars;
	protected PrismComponent mc;
	protected NondetModel mdp;
	
	/**
	 * Returns a deep copy of this
	 */
	public abstract AbstractLiteral clone();

	/**
	 * Performs an intersection between the set of states  and mec if this is an INF or a FIN
	 * @param mec the set of states the literal should be intersected with
	 */
	public abstract void cutWithMEC(JDDNode mec);
	/**
	 * If literal is Fin(P), then remove P from mecs, and generate the new maximal end components
	 * within mecs without transitions in P
	 * @param mecs the maximal end component, which should be pruned
	 * @return the maximal end components within mecs\P
	 * @throws PrismException
	 */
	public abstract List<JDDNode> cutMEC(JDDNode mecs) throws PrismException;

	/**
	 * Compute the MECs within a given state space
	 * @param restrict the state space within the MECs are collected
	 * @param accept collect only MECs, that have states in accept. If null, all MECs are collected
	 * @return List of MECs 
	 * @throws PrismException
	 */
	protected List<JDDNode> getMECs(JDDNode restrict, JDDNode accept) throws PrismException 
	{
		ECComputer ecComputer = ECComputer.createECComputer(mc, mdp);
		ecComputer.computeMECStates(restrict, accept);
		return ecComputer.getMECStates();
		
	}

	public AbstractLiteral(int index, boolean negated) 
	{
		this.index = index;
		kind = negated ? Kind.NEGSLACKVAR : Kind.SLACKVAR;
	}
	
	public AbstractLiteral(JDDNode states, boolean inf, PrismComponent mc, NondetModel mdp)
	{
		this.states = states;
		kind = inf ? Kind.INF : Kind.FIN;
		this.mc = mc;
		this.mdp = mdp;
	}
	
	public AbstractLiteral(AcceptanceGenericDD acc, PrismComponent mc, NondetModel mdp) throws PrismException
	{
		switch(acc.getKind()) {
		case FIN:
			kind = Kind.FIN;
			states = acc.getStates();
			break;
		case INF:
			kind = Kind.INF;
			states = acc.getStates();
			break;
		default:
			throw new PrismException("Not implemented");
		}
		this.mc = mc;
		this.mdp = mdp;
	}

	/**
	 * If literal caused an assignment to this literal, then this assignment will be undone
	 * @param literal the literal that caused an assignment
	 * @param value the value for literal during the original assignment
	 * @param mec the original mec during the assignment
	 * @return true, if the value of this has changed, false otherwise
	 */
	public boolean backtrackFinValue(AbstractLiteral literal, boolean value, JDDNode mec)
	{
		if(!isUnassigned()) {
			switch(getKind()) {
				case FIN:
					if(value) {
						JDDNode lastStates = originalStates.pop();
						JDD.Deref(states);
						states = lastStates;
						if(Utils.isSubsetEq(getStates(), literal.getStates())) {
							setUnassigned();
							return true;
						} else {
							return false;
						}
					} else {
						if(Utils.isSubsetEq(literal.getStates(), originalStates.peek())) {
							JDD.Deref(states);
							states = originalStates.pop();
							setUnassigned();
							return true;
						} else {
							return false;
						}
					}
				case INF:
					if(value) {
						// TODO: Should be included
						// JDD.Deref(states);
						states = originalStates.pop();
						if(Utils.isSubsetEq(getStates(), literal.getStates())) {
							setUnassigned();
							return true;
						} else {
							return false;
						}
					} else {
						if(Utils.isSubsetEq(literal.getStates(), originalStates.peek())) {
							// TODO: Should be included
							// JDD.Deref(states);
							states = originalStates.pop();
							setUnassigned();
							return true;
						} else {
							return false;
						}
						
					}
				default:
					return false;
			}
		} else {
			return false;
		}
	}
	
	public boolean backtrackSlackVarValue(AbstractLiteral literal, boolean value, JDDNode mec)
	{
		switch(getKind()) {
			case SLACKVAR:
				if(getIndex() == literal.getIndex()) {
					if(value) {
						setUnassigned();
						return true;
					} else {
						setUnassigned();
						return true;
					}
				}
				return false;
			case NEGSLACKVAR:
				if(getIndex() == literal.getIndex()) {
					if(value) {
						setUnassigned();
						return true;
					} else {
						setUnassigned();
						return true;
					}
				}
				return false;
			default:
				return false;
		}
	}
	
	public boolean backtrackNegSlackVarValue(AbstractLiteral literal, boolean value, JDDNode mec)
	{
		return backtrackSlackVarValue(literal.complement(), !value, mec);
	}
	
	public boolean backtrack(AbstractLiteral literal, boolean value, JDDNode mec)
	{
		switch(literal.getKind()) {
		case SLACKVAR: return backtrackSlackVarValue(literal, value, mec);
		case NEGSLACKVAR: return backtrackNegSlackVarValue(literal, value, mec);
		case FIN: return backtrackFinValue(literal, value, mec);
		default:
			return false;
		}
	}
	
	
	public boolean equals(Object o)
	{
		if(!(o instanceof AbstractLiteral)) {
			return false;
		}
		
		AbstractLiteral other = (AbstractLiteral)o;
		if(other.getKind() != getKind()) {
			return false;
		}
		
		switch(getKind()) {
		case INF:
		case FIN:
			return getStates().equals(other.getStates());
		case SLACKVAR:
		case NEGSLACKVAR:
			return getIndex() == other.getIndex();
		}
		return false;
	}
	
	public int hashCode() 
	{
		switch(getKind()) {
		case SLACKVAR:
			return getIndex();
		case NEGSLACKVAR:
			return ~getIndex();
		case INF:
			return getStates().hashCode();
		case FIN:
			return ~getStates().hashCode();
		}
		
		return 0;
	}
	
	/**
	 * signals the value of the literal
	 * @return true, if this is false
	 */
	public boolean isFalse()
	{
		return value == Assignment.FALSE;
	}
	
	/**
	 * sets this to false
	 */
	public void setFalse()
	{
		value = Assignment.FALSE;
	}

	/**
	 * signals the value of the literal
	 * @return true, if this is true
	 */
	public boolean isTrue()
	{
		return value == Assignment.TRUE;
	}
	
	/**
	 * sets this to true
	 */
	public void setTrue()
	{
		value = Assignment.TRUE;
	}
	
	/**
	 * signals the value of the literal
	 * @return true, if this is unassigned
	 */
	public boolean isUnassigned()
	{
		return value == Assignment.UNASSIGNED;
	}
	
	/**
	 * sets this to unassigned
	 */
	public void setUnassigned()
	{
		value = Assignment.UNASSIGNED;
	}

	public int getIndex() {
		return index;
	}
	
	public Kind getKind() 
	{
		return kind;
	}
	
	public JDDNode getStates()
	{
		return states;
	}
	
	public boolean assignSlackVarValue(AbstractLiteral literal, boolean value)
	{
		switch(getKind()) {
			case SLACKVAR:
				if(getIndex() == literal.getIndex()) {
					if(value) {
						setTrue();
						return true;
					} else {
						setFalse();
						return true;
					}
				}
				return false;
			case NEGSLACKVAR:
				if(getIndex() == literal.getIndex()) {
					if(value) {
						setFalse();
						return true;
					} else {
						setTrue();
						return true;
					}
				}
				return false;
			default:
				return false;
		}
	}
	
	public abstract AbstractLiteral complement();
	
	public boolean assignNegSlackVarValue(AbstractLiteral literal, boolean value)
	{
		return assignSlackVarValue(literal.complement(), !value);
	}
	
	public boolean assignFinValue(AbstractLiteral literal, boolean value, JDDNode mec)
	{
		switch(getKind()) {
			case FIN:
				if(value) {
					originalStates.push(getStates().copy());
					if(Utils.isSubsetEq(getStates(), literal.getStates())) {
						setTrue();
						return true;
					} else {
						// TODO: Actually, states should be without .copy()...
						states = JDD.And(states,
								JDD.Not(literal.getStates().copy()));
						//states = JDD.And(states.copy(),
						//		JDD.Not(literal.getStates().copy()));
						return false;
					}
				} else {
					if(Utils.isSubsetEq(literal.getStates(), getStates())) {
						originalStates.push(getStates().copy());
						setFalse();
						return true;
					} else {
						return false;
					}
					
				}
			case INF:
				if(value) {
					originalStates.push(getStates().copy());
					if(Utils.isSubsetEq(getStates(), literal.getStates())) {
						setFalse();
						return true;
					} else {
						states = JDD.And(states,
								JDD.Not(literal.getStates().copy()));
						return false;
					}
				} else {
					if(Utils.isSubsetEq(literal.getStates(), getStates())) {
						originalStates.push(getStates().copy());
						setTrue();
						return true;
					} else {
						return false;
					}
					
				}
			default:
				return false;
		}
	}
	
	/**
	 * Assigns the value to this, if it literal agrees with this, i.e.,
	 * they have corresponding kinds
	 * (e.g. SLACKVAR and SLACKVAR or FIN and INF with corresponding states sets)
	 * 
	 * @param literal the literal to which this should be compared
	 * @param value signs, whether true or false should be assigned
	 * @return true, if an assignment happened, false otherwise
	 * @throws PrismException 
	 */
	public boolean assignValue(AbstractLiteral literal, boolean value, JDDNode mec) 
	{
		switch(literal.getKind()) {
		case FIN:
			return assignFinValue(literal, value, mec);
		case SLACKVAR:
			return assignSlackVarValue(literal, value);
		case NEGSLACKVAR:
			return assignNegSlackVarValue(literal, value);
		default:
			return false;
		}
	}
	
	//private String assignmentToString()
	//{
	//	switch(value) {
	//	case TRUE: return "T";
	//	case FALSE: return "F";
	//	case UNASSIGNED: return "U";
	//	}
	//	return null;
	//}
	
	public String toString()
	{
		switch(kind) {
		case SLACKVAR: return "s" + index;
		case NEGSLACKVAR: return "!s" + index;
		case INF: return "Inf(" + getStates() + ")";
		case FIN: return "Fin(" + getStates() + ")";
		default:
			throw new NoSuchElementException();
		}
	}

	@Override
	public int compareTo(AbstractLiteral other) 
	{
		if(getKind() == Kind.SLACKVAR) {
			if(other.getKind() == Kind.SLACKVAR) {
				return Integer.valueOf(getIndex()).compareTo(Integer.valueOf(other.getIndex()));
			}
			if(other.getKind() == Kind.NEGSLACKVAR) {
				return getIndex() <= other.getIndex() ? -1 : 1; 
			}
			return -1;
		}

		if(getKind() == Kind.NEGSLACKVAR) {
			if(other.getKind() == Kind.NEGSLACKVAR) {
				return Integer.valueOf(getIndex()).compareTo(Integer.valueOf(other.getIndex()));
			}
			if(other.getKind() == Kind.SLACKVAR) {
				return other.getIndex() <= getIndex() ? 1 : -1; 
			}
			return -1;
		}

		if(getKind() == Kind.FIN) {
			if(other.getKind() == Kind.SLACKVAR ||
					other.getKind() == Kind.NEGSLACKVAR) {
				return 1;
			}
			
			if(other.getKind() == Kind.FIN) {
				if(getStates().equals(other.getStates())) {
					// both literals equals
					return 0;
				}

				if(JDD.GetNumMinterms(getStates(), numVars) >
					JDD.GetNumMinterms(other.getStates(), numVars)) {
					return -1;
				}
				if(JDD.GetNumMinterms(getStates(), numVars) <
					JDD.GetNumMinterms(other.getStates(), numVars)) {
					return 1;
				}

				// Just do some stuff to get a linear order ;-)
				return Long.compare(getStates().ptr(), other.getStates().ptr());
			}
			// Other literal is a INF literal
			return -1;
		}

		if(getKind() == Kind.INF) {
			if(other.getKind() == Kind.INF) {
				if(getStates().equals(other.getStates())) {
					// both literals equals
					return 0;
				}

				if(JDD.GetNumMinterms(getStates(), numVars) >
					JDD.GetNumMinterms(other.getStates(), numVars)) {
					return -1;
				}
				if(JDD.GetNumMinterms(getStates(), numVars) <
					JDD.GetNumMinterms(other.getStates(), numVars)) {
					return 1;
				}

				// Just do some stuff to get a linear order ;-)
				return Long.compare(getStates().ptr(), other.getStates().ptr());
			}
			// Other literal is a INF literal
			return 1;
		}
		return 0;
	}
	
	public void setStates(JDDNode states)
	{
		this.states = states;
	}

	/**
	 * Derefs a state JDD node if existent
	 */
	public void deref()
	{
		if(getKind() == Kind.INF || getKind() == Kind.FIN) {
			JDD.Deref(getStates());
			derefStack();
		}
	}
	
	public void derefStack()
	{
		originalStates.
			stream().
			forEach(x -> JDD.Deref(x));
	}

	public String toVerboseString() {
		String result;
		switch(kind) {
		case FIN:
			result = "Fin(" + getStates() + ")";
			break;
		case INF:
			result = "Inf(" + getStates() + ")";
			break;
		case NEGSLACKVAR:
			result = "!s" + getIndex();
			break;
		case SLACKVAR:
			result = "s" + getIndex();
			break;
		default:
			result = null;
		}

		if(isUnassigned()) {
			result += "U";
		}
		if(isTrue()) {
			result += "T";
		}
		if(isFalse()) {
			result += "F";
		}
		return result;
		
	}
}
