package symbolic.satsolving;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import acceptance.AcceptanceGeneric.ElementType;
import acceptance.AcceptanceGenericDD;
import jdd.JDD;
import jdd.JDDNode;
import prism.ECComputer;
import prism.NondetModel;
import prism.NondetModelChecker;
import prism.PrismException;
import symbolic.satsolving.AbstractLiteral.Kind;

	
public abstract class CNF extends LinkedList<Clause> {
	/**
	 *  
	 */
	private static final long serialVersionUID = 1L;

	public enum Assignment {
		TRUE,
		FALSE,
		UNASSIGNED
	}	

	protected Assignment value = Assignment.UNASSIGNED;
	/**
	 * Store acceptance conditions to the according slack variables.
	 */
	protected static Map<AcceptanceGenericDD,Integer> acc2SlackVar = new LinkedHashMap<>();
	protected NondetModelChecker mc;
	protected NondetModel mdp;
	
	protected LiteralTreeSet aps = new LiteralTreeSet();

	/**
	 * Singleton pattern for acc2SlackVar, i.e., the mapping acceptance condition -> slack variable
	 * @return global mapping from acceptance conditions to slack variables
	 */
	protected static Map<AcceptanceGenericDD,Integer> getAcc2SlackVar()
	{
		return acc2SlackVar;
	}
	
	/**
	 * Search for acceptance condition acc, 
	 * If it is already known, then return the according slack variable.
	 * Otherwise, create a new slack variable for acc and store it
	 * @param acc the acceptance condition, for which a slack var is searched
	 * @return a slack variable for acc
	 * @throws PrismException
	 */
	protected static int findOrAdd(AcceptanceGenericDD acc) throws PrismException
	{
		Map<AcceptanceGenericDD,Integer> acc2SlackVar = getAcc2SlackVar();
		//Search for acc in acc2SlackVar
		if(!acc2SlackVar.containsKey(acc)) {
			int max = acc2SlackVar.size();
			
			//Sanity check: every variable should be inserted only once
			if(acc2SlackVar.containsValue(max)) {
				throw new PrismException("Integer " + max + " is contained two times in hash table for the Tseytin transformation of " + acc);
			}
			
			//Store new slack variable
			acc2SlackVar.put(acc, max);
		}
		return acc2SlackVar.get(acc);
	}
	
	/**
	 * Compute the MECs within a given state space
	 * @param restrict the state space within the MECs are collected
	 * @param accept collect only MECs, that have states in accept. If null, all MECs are collected
	 * @return List of MECs 
	 * @throws PrismException
	 */
	protected List<JDDNode> getMECs(JDDNode restrict, JDDNode accept) throws PrismException 
	{
		ECComputer ecComputer = ECComputer.createECComputer(mc, mdp);
		return ecComputer.getMECStates();
		
	}
	
	protected abstract AbstractLiteral createLiteral(JDDNode node, boolean inf);
	protected abstract AbstractLiteral createLiteral(int index, boolean negated);
	protected abstract AbstractLiteral createLiteral(AcceptanceGenericDD acc) throws PrismException;

	public abstract List<JDDNode> cutMEC(JDDNode mecs, AbstractLiteral literal) throws PrismException;
	
	public void backtrack(AbstractLiteral literal, boolean value, JDDNode mec)
	{
		for(Clause clause : this) {
			clause.backtrack(literal, value, mec);
			if(clause.isFalse()) {
				this.value = Assignment.FALSE;
			}
			if(this.value != Assignment.FALSE && clause.isUnassigned()) {
				this.value = Assignment.UNASSIGNED;
			}
		}
	}

	/**
	 * Intersect every state set in a FIN or INF in this with mec
	 * @param mec the set of states for the intersection 
	 */
	public void cutWithMEC(JDDNode mec)
	{
		stream().
			forEach(x -> x.cutWithMEC(mec));
		simplify();
	}
	
	/**
	 * Simplifies the formula
	 */
	public void simplify()
	{
		//If this formula is true, do nothing
		if(isTrue()) {
			return;
		}
		
		Iterator<Clause> it = iterator();
		while(it.hasNext()) {
			Clause clause = it.next();
			//Simplify every clause
			clause.simplify();
		}
	}
	
	protected Set<Clause> createClauseSetInf(AcceptanceGenericDD acc)
	{
		Set<Clause> clauses = new LinkedHashSet<>();
		
		Clause clause = new Clause(mc, mdp);

		AbstractLiteral infLit = createLiteral(acc.getStates().copy(), true);
		clause.add(infLit);
		clauses.add(clause);
		aps.add(infLit);
		
		return clauses;
	}
	
	protected Set<Clause> createClauseSetFin(AcceptanceGenericDD acc)
	{
		Set<Clause> clauses = new LinkedHashSet<>();
		
		Clause clause = new Clause(mc, mdp);

		AbstractLiteral finLit = createLiteral(acc.getStates().copy(), false);
		clause.add(finLit);
		clauses.add(clause);
		aps.add(finLit);
		
		return clauses;
	}
	
	protected Set<Clause> createClauseSetDisj(AcceptanceGenericDD acc, Clause clause) throws PrismException
	{
		//Rule for or:
		//s_acc -> (s_left | s_right)
		Set<Clause> clauses = new LinkedHashSet<>();
		
		if(acc.getLeft().getKind() == ElementType.FIN ||
				acc.getLeft().getKind() == ElementType.INF) {
			//Add a FIN or INF directly
			AbstractLiteral lit = createLiteral(acc.getLeft().copy());
			clause.add(lit);
			aps.add(lit);
		} else {
			if(acc.getLeft().getKind() == ElementType.OR) {
				clauses.addAll(createClauseSetDisj(acc.getLeft(), clause));
			} else {
				//Add a boolean combination via findOrAdd
				AbstractLiteral slackLit = createLiteral(findOrAdd(acc.getLeft()), false);
				clause.add(slackLit);
				clauses.addAll(createClauseSet(acc.getLeft()));
				aps.add(slackLit);
			}
		}

		if(acc.getRight().getKind() == ElementType.FIN ||
				acc.getRight().getKind() == ElementType.INF) {
			//Add a FIN or INF directly
			AbstractLiteral lit = createLiteral(acc.getRight().copy());
			clause.add(lit);
			aps.add(lit);
		} else {
			if(acc.getRight().getKind() == ElementType.OR) {
				clauses.addAll(createClauseSetDisj(acc.getRight(), clause));
			} else {
				//Add a boolean combination via findOrAdd
				AbstractLiteral slackLit = createLiteral(findOrAdd(acc.getRight()), false);
				clause.add(slackLit);
				clauses.addAll(createClauseSet(acc.getRight()));
				aps.add(slackLit);
			}
		}
		
		return clauses;
	}
	protected Set<Clause> createClauseSetConj(AcceptanceGenericDD acc, AbstractLiteral negSlackVar) throws PrismException
	{
		//Rule for and:
		//(s_acc -> s_left) & (s_acc -> s_right)
		Set<Clause> clauses = new LinkedHashSet<>();
		
		Clause clauseLeft = new Clause(mc, mdp);
		clauseLeft.add(negSlackVar);

		if(acc.getLeft().getKind() == ElementType.FIN ||
				acc.getLeft().getKind() == ElementType.INF) {
			//Add a FIN or INF directly
			AbstractLiteral lit = createLiteral(acc.getLeft().copy());
			clauseLeft.add(lit);
			aps.add(lit);
			clauses.add(clauseLeft);
		} else {
			AbstractLiteral slackLit = null;
			if(acc.getLeft().getKind() == ElementType.AND) {
				slackLit = negSlackVar.clone();
				clauses.addAll(createClauseSetConj(acc.getLeft(), negSlackVar));
			} else {
				slackLit = createLiteral(findOrAdd(acc.getLeft()), false);
				clauses.addAll(createClauseSet(acc.getLeft()));
				clauseLeft.add(slackLit);
				aps.add(slackLit);
				clauses.add(clauseLeft);
			}
		}
			
		Clause clauseRight = new Clause(mc, mdp);
		clauseRight.add(negSlackVar);

		if(acc.getRight().getKind() == ElementType.FIN ||
				acc.getRight().getKind() == ElementType.INF) {
			//Add a FIN or INF directly
			AbstractLiteral lit = createLiteral(acc.getRight().copy());
			clauseRight.add(lit);
			aps.add(lit);
		} else {
			AbstractLiteral slackLit = null;
			if(acc.getRight().getKind() == ElementType.AND) {
				slackLit = negSlackVar.clone();
				clauses.addAll(createClauseSetConj(acc.getRight(), negSlackVar));
			} else {
				//Add a boolean combination via findOrAdd
				slackLit = createLiteral(findOrAdd(acc.getRight()), false);
				clauses.addAll(createClauseSet(acc.getRight()));
			}
			clauseRight.add(slackLit);
			aps.add(slackLit);
		}
		clauses.add(clauseRight);
		
		return clauses;
	}
	
	
	/**
	 * Applies the Tseytin transformation for the generic acceptance to acc, except adding a unit clause for the top node
	 * @param acc the generic acceptance which should be transformed.
	 * @return set of clauses for acc
	 * @throws PrismException
	 */
	private Set<Clause> createClauseSet(AcceptanceGenericDD acc) throws PrismException
	{
		switch(acc.getKind()) {
		case FIN:
			return createClauseSetFin(acc);
		case INF:
			return createClauseSetInf(acc);
		case AND:
			AbstractLiteral negLiteral = createLiteral(findOrAdd(acc),true);
			return createClauseSetConj(acc, negLiteral);
		case OR:
			negLiteral = createLiteral(findOrAdd(acc),true);
			Clause clause = new Clause(mc, mdp);
			clause.add(negLiteral);
			Set<Clause> clauses = createClauseSetDisj(acc, clause);
			clauses.add(clause);
			return clauses;
		case FALSE:
		case FIN_NOT:
		case INF_NOT:
		case TRUE:
		default:
			throw new PrismException("Tseytin transformation is not yet implemented for FALSE, FIN_NOT, INF_NOT and TRUE");
		}
	}
	
	/**
	 * Constructor for CNF
	 * @param acc the generic acceptance, for which a CNF should be create
	 * @param mdp the MDP, which should be analysed under acc
	 * @param mc the according MDPModelChecker
	 * @throws PrismException
	 */
	public CNF(AcceptanceGenericDD acc, NondetModel mdp, NondetModelChecker mc) throws PrismException
	{
		this(mdp, mc);
		
		addAll(createClauseSet(acc));

		Clause top = new Clause(mc, mdp);
		AbstractLiteral topLit = createLiteral(findOrAdd(acc),false);
		top.add(topLit);
		
		add(top);
		aps.add(topLit);
		unifyAllInfs();
		recalculateAssignment();
	}

	/**
	 * Checks, whether it can be shown be simple transformations, if this is equivalent to true
	 * @return true, if this is equivalent to true
	 */
	public boolean isTrue()
	{
		for(Clause clause : this) {
			if(clause.isFalse()) {
				return false;
			}
			if(!clause.isTrue()) {
				//CNF, that contains only INF as atomic propositions are also true 
				for(AbstractLiteral ap : clause) {
					if(ap.isUnassigned() && ap.getKind() != AbstractLiteral.Kind.INF) {
						return false;
					}
				}
			}
		}
		return true;
	}
	
	/**
	 * Checks, whether it can be shown be simple transformations, if this is equivalent to false
	 * @return true, if this is equivalent to false
	 */
	public boolean isFalse()
	{
		recalculateAssignment();
		return value == Assignment.FALSE;
	}
	
	@Override
	/**
	 * Returns a textual representation of this.
	 */
	public String toString() 
	{
		if(isTrue()) {
			return "true";
		}
		if(isFalse()) {
			return "false";
		}
		String result = "";
		Iterator<Clause> it = iterator();
		if(!it.hasNext()) {
			return "true";
		}
		Clause clause = it.next();
		while(clause.isTrue() && it.hasNext()) {
			clause = it.next();
		}
		if(clause.isTrue() && !it.hasNext()) {
			return "true";
		}
		result = "(" + clause.toString() + ")";
		while(it.hasNext()) {
			clause = it.next();
			if(!clause.isTrue()) {
				result += " & (" + clause.toString() + ")";
			}
		}
		return result;
	}
	
	/**
	 * Chooses an unassigned branching literal 
	 * (a FIN, a slack variable, or a negated slack variable)
	 * @return branching literal
	 */
	public AbstractLiteral chooseAP() 
	{
		AbstractLiteral chosenAP = aps.pollFirst();
		if(chosenAP.getKind() == Kind.INF) {
			return null;
		}
		return chosenAP.clone();
	}
	
	/**
	 * Constructor for CNF
	 * @param mdp the MDP, that should be analysed
	 * @param mc the according MDP model checker
	 */
	protected CNF(NondetModel mdp, NondetModelChecker mc)
	{
		this.mdp = mdp;
		this.mc = mc;
	}
	
	/**
	 * Searchs for a pure literal in this. If found, it is return, otherwise null is returned
	 * @return a pure literal, if existent. Null, otherwise
	 */
	public AbstractLiteral findPureLiteral() 
	{
		//Set of potential pure literals
		LinkedHashSet<AbstractLiteral> pureLiterals = new LinkedHashSet<AbstractLiteral>();
		//Set of literals, where a positive and a negative literal has been found
		LinkedHashSet<AbstractLiteral> removedLiterals = new LinkedHashSet<AbstractLiteral>();
		for(Clause clause : this) {
			if(clause.isUnassigned()) {
				for(AbstractLiteral ap: clause) {
					if(ap.isUnassigned() && !removedLiterals.contains(ap)) {
						if(ap.getKind() == Kind.SLACKVAR || ap.getKind() == Kind.NEGSLACKVAR) {
							//Negate variable
							AbstractLiteral negVar = ap.complement();
							//Search for negated variable
							if(pureLiterals.contains(negVar)) {
								//If found, remove it from potential pure literals and 
								//add positive and negative literal to removedLiterals
								pureLiterals.remove(negVar);
								removedLiterals.add(negVar);
								removedLiterals.add(ap);
							} else {
								//If not found, ap is a possible pure literal
								pureLiterals.add(ap);
							}
						}
					}
				}
			}
		}
		//Return a pure literal, if there was one found
		if(!pureLiterals.isEmpty()) {
			Iterator<AbstractLiteral> it = pureLiterals.iterator();
			if(it.hasNext()) {
				return it.next().clone();
			}
		}
		return null;
	}
	
	/**
	 * Performs the DPLL algorithm on this (and the maximal end components mecs)
	 * @param mec the maximal end component, for which a model should be searched for
	 * @return the end components inducing a model for this
	 * @throws PrismException
	 */
	public JDDNode dpll(JDDNode mec)  throws PrismException
	{
		
		//If mecs is empty, this cannot be satisfied 
		if(mec.equals(JDD.ZERO)) {
			return null;
		}
		
		//Remove INF(emptyset) and FIN(emptyset)
		recalculateAssignment();
		//If this is true, then mecs is trivially a model
		if(isTrue()) {
			return mec.copy();
		}
		
		//If this is false, then there is trivially no model to this
		if(isFalse()) {
			return null;
		}
		
		//Check for Unit-Clause
		for(Clause clause : this) {
			AbstractLiteral unitAP = clause.isUnitClause();
			if(unitAP != null) {
				aps.remove(unitAP);
				List<JDDNode> subMecs = cutMEC(mec, unitAP);
				for(JDDNode subMec : subMecs) {
					assignValue(unitAP, true, subMec);
					JDDNode satisfyingEC = dpll(subMec);
					if(satisfyingEC != null) {
						JDD.DerefAll(subMecs);
						return satisfyingEC;
					}
					backtrack(unitAP, true, subMec);
				}
				// Clean up subMECs
				JDD.DerefAll(subMecs);
				//If no satisfying EC has been found
				//this cannot be satisfied
				aps.add(unitAP);
				return null;
			}
		}

		//Search for a pure literal
		AbstractLiteral pureLiteral = findPureLiteral();
		if(pureLiteral != null) {
				aps.remove(pureLiteral);
				List<JDDNode> subMecs = cutMEC(mec, pureLiteral);
				for(JDDNode subMec : subMecs) {
					assignValue(pureLiteral, true, subMec);
					JDDNode satisfyingEC = dpll(subMec);
					if(satisfyingEC != null) {
						JDD.DerefAll(subMecs);
						return satisfyingEC;
					}
					backtrack(pureLiteral, true, subMec);
				}
				// Clean up all subMecs
				JDD.DerefAll(subMecs);
				//If no satisfying EC has been found
				//this cannot be satisfied
				aps.add(pureLiteral);
				return null;
		}
		
		//Branch over a literal
		AbstractLiteral chosenLiteral = chooseAP();
		if(chosenLiteral.getKind() != AbstractLiteral.Kind.FIN &&
				chosenLiteral.getKind() != AbstractLiteral.Kind.SLACKVAR &&
				chosenLiteral.getKind() != AbstractLiteral.Kind.NEGSLACKVAR) {
			throw new PrismException("Cannot handle chosen INF Literals");
		}

		List<JDDNode> subMecs = cutMEC(mec, chosenLiteral);
		//Assign true to chosenLiteral
		for(JDDNode subMec : subMecs) {
			assignValue(chosenLiteral, true, subMec);
			JDDNode leftEC = dpll(subMec);
			if(leftEC != null) {
				JDD.DerefAll(subMecs);
				return leftEC;
			}
			backtrack(chosenLiteral, true, subMec);
		}
		JDD.DerefAll(subMecs);

		//Assign false to chosenLiteral
		assignValue(chosenLiteral, false, mec);
		JDDNode rightEC = dpll(mec);
		if(rightEC != null) {
			return rightEC;
		}
		backtrack(chosenLiteral, false, mec);
		aps.add(chosenLiteral);
		return null;
	}
	
	
	
	
	///**
	// * Joins a list of MECs to a BitSet
	// * @param mecs list of MECs
	// * @return joined set of states in mecs
	// */
	//private BitSet joinMECs(List<BitSet> mecs) 
	//{
	//	BitSet joined = new BitSet();
	//	for(BitSet mec : mecs) {
	//		joined.or(mec);
	//	} 
	//	return joined;
	//}
	
	/**
	 * Does one branching step, i.e., sets literal to value and adopts this
	 * @param literal the branching literal
	 * @param value the value for literal
	 */
	public void assignValue(AbstractLiteral literal, boolean value, JDDNode mec)
	{
		for(Iterator<Clause> it = iterator(); it.hasNext(); ) {
			Clause clause = it.next();
			if(clause.isUnassigned()) {
				for(AbstractLiteral litClause : clause) {
					if(litClause.isUnassigned()) {
						boolean isAssigned = litClause.assignValue(literal, value, mec);
						if(isAssigned && litClause.isTrue()) {
							clause.setTrue();
						}
					}
				}
				clause.recalculateAssignment();
			}
		}

		recalculateAssignment();
	}
	
	/**
	 * Reduce two or more Infs in every clause 
	 * two one Inf (by union of the states)
	 */
	protected void unifyAllInfs()
	{
		for(Clause clause : this){
			clause.unifyAllInfs();
		}
	}
	
	/**
	 * Checks, whether this can be transformed to true, or false via simple transformation steps
	 */
	protected void recalculateAssignment()
	{
		boolean foundUnassigned = false;
		//Search for false and unassigned clauses
		for(Clause clause : this) {
			clause.recalculateAssignment();
			//If this contains a false clause, this is false
			if(clause.isFalse()) {
				value = Assignment.FALSE;
				return;
			}
			//If this contains an unassigned clause (and no false clause),
			//then this is unassigned
			if(clause.isUnassigned()) {
			  foundUnassigned = true;
			}
		}
		value = foundUnassigned ? Assignment.UNASSIGNED : Assignment.TRUE;
	}
	
	protected Clause getFalseClause()
	{
		recalculateAssignment();
		assert isFalse();
		return stream().
				filter(clause -> clause.isFalse()).
				findAny().
				get();
	}
	
	//Debugging function
	//Delete, if Debugging is done
	public boolean containsInfWithEmptySet()
	{
		for(Clause clause : this) {
			for(AbstractLiteral literal : clause) {
				if(literal.getKind() == Kind.INF && literal.getStates().equals(JDD.ZERO) && literal.isUnassigned()) {
					return true;
				}
			}
		}
		return false;
	}
	
	/**
	 * Removes clauses, that contains a true literal (e.g. Fin(nothing)) and removes false literals
	 */
	public void removeTrueFalse()
	{
		recalculateAssignment();
		for(Iterator<Clause> it = iterator(); it.hasNext();) {
			Clause clause = it.next();
			if(clause.isTrue()) {
				clause.deref();
				it.remove();
			}
		}
		
		for(Clause clause : this) {
			clause.removeFalse();
		}
	}
	
	public void removeDuplicates()
	{
		for(Clause clause : this) {
			clause.removeDuplicates();
		}
		
		Set<Clause> clauses = new HashSet<Clause>();
		for(Iterator<Clause> it = iterator(); it.hasNext();) {
			Clause clause = it.next();
			if(clauses.contains(clause)) {
				clause.deref();
				it.remove();
			} else {
				clauses.add(clause);
			}
		}
	}
	
	/**
	 * Derefs all literals with state sets
	 */
	public void deref()
	{
		stream().
		forEach(x -> x.deref());
	}

	public String toVerboseString()
	{
		String result;
		if(size() > 0) {
			Iterator<Clause> it = iterator();
			result = "(" + it.next().toVerboseString() + ")";
			while(it.hasNext()) {
				result += "&" + "(" + it.next().toVerboseString() + ")";
			}
		} else {
			result = "true";
		}
		return result;
	}
}
