package symbolic.satsolving;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import java.util.stream.Collectors;

import jdd.JDD;
import jdd.JDDNode;
import prism.NondetModel;
import prism.PrismComponent;

class Clause extends LinkedList<AbstractLiteral> 
{
		
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	protected PrismComponent mc;
	protected NondetModel mdp;

	public enum Assignment {
			TRUE,
			FALSE,
			UNASSIGNED
	}
	
	/**
	 * signs, whether the clause is false, true or unassigned
	 */
	protected Assignment value = Assignment.UNASSIGNED;
	
	/**
	 * Checks, whether this is equivalent to true after simple transformations 
	 * @return true, if this is equivalent to true
	 */
	public boolean isTrue() {
		recalculateAssignment();
		return value == Assignment.TRUE;
	}
	
	/**
	 * Sets this to true
	 */
	public void setTrue() {
		value = Assignment.TRUE;
	}
	
	
	/**
	 * Performs an intersection between the set of states in every FIN and INF literal and mec
	 * @param mec the set of states the literals should be intersected with
	 */
	public void cutWithMEC(JDDNode mec)
	{
		//Go through every literal, intersect, if it is of the form INF or FIN
		stream().
			forEach(x -> x.cutWithMEC(mec));
		simplify();
	}
	
	/**
	 * If two Infs-Literal INF(P1) and INF(P2) appears in a clause, remove both and
	 * add INF(P1 \cup P2)
	 */
	public void unifyAllInfs()
	{
		JDDNode inf = JDD.ZERO.copy();
		Iterator<AbstractLiteral> it = iterator();
		boolean stateAcc = true;
		NondetModel model = null;
		while(it.hasNext()) {
			AbstractLiteral ap = it.next();

			if(ap.getKind() == AbstractLiteral.Kind.INF && ap.isUnassigned()) {
				if(ap instanceof TransLiteral) {
					stateAcc = false;
					model = ((TransLiteral) ap).getModel();
				}
				//Store the states in inf and
				//remove the ap
				inf = JDD.Or(inf, ap.getStates());

				it.remove();
			}
		}
		
		if(inf.equals(JDD.ZERO)) {
			JDD.Deref(inf);
		} else {
			if(stateAcc) {
				add(new StateLiteral(inf, true, mc, model));
			} else {
				add(new TransLiteral(inf, true, mc, model));
			}
		}
	}
	
	/**
	 * Performs simple transformation preserving equivalence
	 */
	public void simplify()
	{
		if(value != Assignment.TRUE && value != Assignment.FALSE) {
			//Search for empty Fin
			Iterator<AbstractLiteral> it = iterator();
			while(it.hasNext()) {
				AbstractLiteral ap = it.next();
				if(ap.getKind() == AbstractLiteral.Kind.FIN && ap.getStates().equals(JDD.ZERO)) {
					ap.setTrue();
					value = Assignment.TRUE;
					return;
				}
			}
		
			//TODO: Set clause to true, if the are aps of the form p and not p (and analogously Fin and Inf)
		
			//TODO: Remove Fin(S'), if there is a Fin(S) with S \subseteq S'
		
		}
		
	}
	
	/**
	 * Returns a textual representation of this clause
	 */
	public String toString() 
	{
		if(value == Assignment.TRUE) {
			return "true";
		}
		
		if(value == Assignment.FALSE) {
			return "false";
		}
		
		String result = "";
		Iterator<AbstractLiteral> it = iterator();
		if(!it.hasNext()) {
			return "false";
		}
		AbstractLiteral literal = it.next();;
		while(literal.isFalse() && it.hasNext()) {
			literal = it.next();
		}
		if(literal.isFalse() && !it.hasNext()) {
			return "false";
		}
		result = literal.toString();
		while(it.hasNext()) {
			literal = it.next();
			if(!literal.isTrue()) {
				result += " | " + literal.toString();
			}
		}
		return result;
	}

	/**
	 * Performs simple transformations preserving equivalence, and
	 * subsequently decides, whether this is false, true or unassigned
	 */
	public void recalculateAssignment() 
	{
		simplify();
		//An empty clause is false
		if(size() == 0) {
			value = Assignment.FALSE;
			return;
		}
		//A true literal implies that the clause is true
		for(AbstractLiteral literal : this) {
			if(literal.isTrue()) {
				value = Assignment.TRUE;
				return;
			}
		}
		//Search for unassigned literals
		Set<AbstractLiteral> unassignedList = stream().filter(literal -> literal.isUnassigned()).collect(Collectors.toSet());
		if(unassignedList.isEmpty()) {
			//If there are no unassigned literals, and the clause is not true, it is false
			value = Assignment.FALSE;
		} else {
			//If there is an unassigned literal (and no true literals)
			//then a clause is unassigned
			value = Assignment.UNASSIGNED;
		}
	}

	/**
	 * Checks, whether this is equivalent to false after simple transformations 
	 * @return true, if this is equivalent to false
	 */
	public boolean isFalse() {
		recalculateAssignment();
		return value == Assignment.FALSE;
	}
	
	/**
	 * Checks, whether this is a unit clause 
	 * Note that, (INF(P) | FIN(Q)), (INF(P) | s), (INF(P) | !s) are not unit clauses
	 * @return the literal in the unit, if this is a unit clause, null otherwise
	 */
	public AbstractLiteral isUnitClause()
	{
		if(value != Assignment.UNASSIGNED) {
			return null;
		}
		
		Set<AbstractLiteral> unassignedList = stream().
				filter(literal -> literal.isUnassigned()).
				collect(Collectors.toSet());
		if(unassignedList.size() != 1) {
			return null;
		}
		Iterator<AbstractLiteral> it = unassignedList.iterator();
		it.hasNext();
		AbstractLiteral unitLit = it.next();
		if(unitLit.getKind() != AbstractLiteral.Kind.INF) {
			return unitLit.clone();
		}
		return null;
	}
	
	public void backtrack(AbstractLiteral literal, boolean value, JDDNode mec)
	{
		for(AbstractLiteral clauseLit : this) {
			clauseLit.backtrack(literal, value, mec);
			if(clauseLit.isTrue()) {
				this.value = Assignment.TRUE;
			}
			if(this.value == Assignment.FALSE && clauseLit.isUnassigned()) {
				this.value = Assignment.UNASSIGNED;
			}
		}
	}
	
	public Clause(PrismComponent mc, NondetModel mdp)
	{
		this.mc = mc;
		this.mdp = mdp;
	}
	
	/**
	 * returns a deep copy of this
	 */
	public Clause clone()
	{
		Clause clause = new Clause(mc, mdp);
		for(AbstractLiteral literal : this) {
			clause.add(literal.clone());
		}
		return clause;
	}
	
	/**
	 * Checks, whether this is unassigned
	 * @return true, if this is unassigned
	 */
	public boolean isUnassigned()
	{
		return value==Assignment.UNASSIGNED;
	}

	/**
	 * Remove all false Literals in a clause
	 */
	public void removeFalse()
	{
		for(Iterator<AbstractLiteral> it = iterator(); it.hasNext(); ) {
			AbstractLiteral lit = it.next();
			if(lit.isFalse()) {
				lit.deref();
				it.remove();
			}
		}
	}
	
	public void removeDuplicates()
	{
		Set<AbstractLiteral> literals = new HashSet<AbstractLiteral>();
		for(Iterator<AbstractLiteral> it = iterator(); it.hasNext();) {
			AbstractLiteral lit = it.next();
			if(literals.contains(lit)) {
				lit.deref();
				it.remove();
			} else {
				literals.add(lit);
			}
		}
	}

	public String toVerboseString()
	{
		String result;
		if(size() > 0) {
			Iterator<AbstractLiteral> it = iterator();
			result = it.next().toVerboseString();
			while(it.hasNext()) {
				result += "|" + it.next().toVerboseString();
			}
		} else {
			result = "false";
		}
		return result;
	}
	
	public void deref()
	{
		stream().
			forEach(x -> x.deref());
	}

	public void derefStack()
	{
		stream().
			forEach(x -> x.derefStack());
	}
	//@Override
	//public int compareTo(Clause o) {
	//	if(size() < o.size()) {
	//		return -1;
	//	}

	//	if(size() > o.size()) {
	//		return 1;
	//	}
	//	
	//	Iterator<Literal> otherIt = o.iterator();
	//	
	//	for(Iterator<Literal> it = iterator(); it.hasNext();) {
	//		Literal lit = it.next();
	//		Literal otherLit = otherIt.next();
	//		
	//	}
	//}
}
